# -*- coding: utf-8 -*-
import pytest

from ddb.binary.binary import DefaultBinary, Binary
from ddb.feature.docker.binaries import DockerBinary
from ddb.registry import Registry, DefaultRegistryObject, RegistryOrderedSet


class Dummy(DefaultRegistryObject):
    pass


class Another(DefaultRegistryObject):
    pass


class HashbableDummy(DefaultRegistryObject):
    def __init__(self, name, a, b):
        super().__init__(name)
        self.a = a
        self.b = b

    def __eq__(self, other):
        if not isinstance(other, HashbableDummy):
            return False
        return self.a == other.a and self.b == other.b

    def __hash__(self):
        return hash(self.a) ^ hash(self.b)


class NoName:
    pass


class TestRegistry:
    def test_empty(self):
        r = Registry(Dummy, "Dummy")

        assert r.all() == ()

    def test_register(self):
        r = Registry(Dummy, "Dummy")

        d = Dummy("dummy")
        r.register(d)

        assert r.get("dummy") == d
        assert r.all() == (d,)

    def test_register_and_unregister(self):
        r = Registry(Dummy, "Dummy")

        d = Dummy("dummy")
        r.register(d)

        assert r.get("dummy") == d
        assert r.all() == (d,)

        r.unregister("dummy")
        with pytest.raises(ValueError) as e:
            r.get("dummy")
        assert str(e.value) == "Dummy name \"dummy\" is not registered"
        assert not r.has("dummy")
        assert r.all() == ()

    def test_register_alternative_name(self):
        r = Registry(Dummy, "Dummy")

        d = Dummy("dummy")
        r.register(d, "alt")

        with pytest.raises(ValueError) as e:
            r.get("dummy")
        assert str(e.value) == "Dummy name \"dummy\" is not registered"

        assert r.get("alt") == d
        assert r.all() == (d,)

    def test_cant_register_many_times(self):
        r = Registry(Dummy, "Dummy")

        d = Dummy("dummy")
        r.register(d)

        with pytest.raises(ValueError) as e:
            r.register(d)
        assert str(e.value) == "Dummy name \"dummy\" is already registered"

    def test_unregistered(self):
        r = Registry(Dummy, "Dummy")

        with pytest.raises(ValueError) as e:
            r.get("dummy")
        assert str(e.value) == "Dummy name \"dummy\" is not registered"

    def test_register_another_class(self):
        r = Registry(Dummy, "Dummy")

        with pytest.raises(ValueError) as e:
            r.register(Another("testing"))
        assert str(e.value) == "Dummy name \"testing\" should be an instance of Dummy"

    def test_register_invalid_instance(self):
        r = Registry(Dummy, "Dummy")

        n = NoName()

        with pytest.raises(ValueError) as e:
            r.register(n, "no-name")
        assert str(e.value) == "Dummy name \"no-name\" should be an instance of Dummy"

    def test_register_no_name_class_missing_name(self):
        r = Registry(NoName, "NoName")

        n = NoName()

        with pytest.raises(ValueError) as e:
            r.register(n)
        assert str(e.value) == "Name should be provided to register this kind of object"


class TestRegistrySet:
    def test_empty(self):
        r = RegistryOrderedSet(HashbableDummy, "HashbableDummy")

        assert len(r.all()) == 0

    def test_register(self):
        r = RegistryOrderedSet(HashbableDummy, "HashbableDummy")

        d = HashbableDummy("dummy", "1", "2")
        r.register(d)

        assert r.get("dummy") == [d]

    def test_register_and_unregister(self):
        r = RegistryOrderedSet(HashbableDummy, "HashbableDummy")

        d = HashbableDummy("dummy", "1", "2")
        r.register(d)

        assert r.get("dummy") == {d}
        assert r.all() == (d,)

        r.unregister("dummy")
        with pytest.raises(ValueError) as e:
            r.get("dummy")
        assert str(e.value) == "HashbableDummy name \"dummy\" is not registered"
        assert not r.has("dummy")
        assert r.all() == ()

    def test_register_alternative_name(self):
        r = RegistryOrderedSet(HashbableDummy, "HashbableDummy")

        d = HashbableDummy("dummy", "1", "2")
        r.register(d, "alt")

        with pytest.raises(ValueError) as e:
            r.get("dummy")
        assert str(e.value) == "HashbableDummy name \"dummy\" is not registered"

        assert r.get("alt") == {d}

    def test_skip_allready_registered_same_instance(self):
        r = RegistryOrderedSet(HashbableDummy, "HashbableDummy")

        d = HashbableDummy("dummy", "1", "2")
        r.register(d)

        d2 = HashbableDummy("dummy", "1", "2")
        with pytest.raises(ValueError) as e:
            r.register(d2)
        assert str(e.value) == "This HashbableDummy is already registered for name \"dummy\""

        d3 = HashbableDummy("dummy", "1", "3")
        r.register(d3)

        assert r.get("dummy") == {d, d3}
        assert r.all() == (d, d3)

    def test_unregistered(self):
        r = RegistryOrderedSet(HashbableDummy, "HashbableDummy")

        with pytest.raises(ValueError) as e:
            r.get("dummy")
        assert str(e.value) == "HashbableDummy name \"dummy\" is not registered"

    def test_register_another_class(self):
        r = RegistryOrderedSet(HashbableDummy, "HashbableDummy")

        with pytest.raises(ValueError) as e:
            r.register(Another("testing"))
        assert str(e.value) == "HashbableDummy name \"testing\" should be an instance of HashbableDummy"

    def test_register_invalid_instance(self):
        r = RegistryOrderedSet(HashbableDummy, "HashbableDummy")

        n = NoName()

        with pytest.raises(ValueError) as e:
            r.register(n, "no-name")
        assert str(e.value) == "HashbableDummy name \"no-name\" should be an instance of HashbableDummy"

    def test_register_no_name_class_missing_name(self):
        r = RegistryOrderedSet(NoName, "NoName")

        n = NoName()

        with pytest.raises(ValueError) as e:
            r.register(n)
        assert str(e.value) == "Name should be provided to register this kind of object"

    def test_binary_ordering(self):
        binary1 = DockerBinary("npm", "node1")
        binary2 = DockerBinary("npm", "node2")
        default_binary = DefaultBinary("npm", ["npm"])
        binary4 = DockerBinary("npm", "node4")
        binary_with_condition = DockerBinary("npm", "node5", condition="some-condition")

        bins = (binary1, binary2, default_binary, binary4, binary_with_condition)

        r = RegistryOrderedSet(Binary, "Binary")
        for bin in bins:
            r.register(bin)

        assert r.all() == bins
        npm_binaries = r.get("npm")

        assert npm_binaries == bins



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema


class SymlinksSchema(FeatureSchema):
    """
    Symlinks schema.
    """
    suffixes = fields.List(fields.String())  # default is build automatically from env value
    includes = fields.List(fields.String())  # default is build automatically from suffixes value
    excludes = fields.List(fields.String())



import os

from ddb.__main__ import main


class TestSymfonyVuejs:
    def test_activate(self, project_loader):
        project_loader("symfony-vuejs")

        main(["configure"])

        main(["activate"])

        assert os.path.exists("docker-compose.override.yml")

        assert sorted(os.listdir(".docker/php/.ca-certificates")) == \
               sorted(("GFI_Informatique_Orléans_Root_CA.crt",
                       "some-cert.crt",
                       "fwca.annuaire.groupe.local.crt"))

        assert os.path.exists(os.path.join(".docker", "db", "Dockerfile"))
        assert os.path.exists(os.path.join(".docker", "php", "Dockerfile"))
        assert os.path.exists(os.path.join(".docker", "php", "fixuid.tar.gz"))
        with open(os.path.join(".docker", "php", "Dockerfile"), "r") as dockerfile:
            data = dockerfile.read()
            assert "COPY .ca-certificates/* /usr/local/share/ca-certificates/" in data
            assert "ADD fixuid.tar.gz /usr/local/bin" in data

        assert os.path.exists(".gitignore")
        with open(".gitignore", "r") as gitignore_file:
            gitignore_lines = gitignore_file.read().splitlines()
            assert gitignore_lines



# -*- coding: utf-8 -*-
from slugify import slugify

from .cache import Cache
from .shelve_cache import ShelveCache
from ..config import config
from ..registry import Registry

caches = Registry(Cache, 'Cache')  # type: Registry[Cache]

global_cache_name = 'global'
requests_cache_name = 'requests'

project_cache_name = 'project'
project_binary_cache_name = 'binary'


def register_project_cache(cache_name):
    """
    Creates a ShelveCache for current project, and register it with given name.
    """
    namespace = [item for item in (project_cache_name, config.paths.project_home, cache_name) if item]
    cache = ShelveCache(slugify('.'.join(namespace), regex_pattern=r'[^-a-z0-9_\.]+'))

    caches.register(cache, cache_name)
    return cache


def register_global_cache(cache_name):
    """
    Creates a ShelveCache shared for all projects, and register it with given name.
    """
    cache = ShelveCache(slugify(cache_name, regex_pattern=r'[^-a-z0-9_\.]+'))
    caches.register(cache, cache_name)
    return cache


def global_cache() -> Cache:
    """
    Get global cache.
    """
    return caches.get(global_cache_name)


def project_cache() -> Cache:
    """
    Get current project cache.
    """
    return caches.get(project_cache_name)


def requests_cache() -> Cache:
    """
    Get current requests cache.
    """
    return caches.get(requests_cache_name)



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod


class Cache(ABC):
    """
    Cache interface
    """

    def __init__(self, namespace: str):
        self._namespace = namespace

    @abstractmethod
    def get(self, key: str, default=None):
        """
        Get a cache entry value
        """

    @abstractmethod
    def keys(self):
        """
        Get all cache keys
        """

    @abstractmethod
    def set(self, key: str, data):
        """
        Set a cache entry value
        """

    @abstractmethod
    def pop(self, key: str):
        """
        Remove a cache entry
        """

    @abstractmethod
    def clear(self):
        """
        Remove all cache entries.
        """

    def close(self):
        """
        Close the cache access.
        """

    def flush(self):
        """
        Synchronize the cache with persistent system.
        """

    def __contains__(self, key):
        """
        Check if key exists.
        """



# -*- coding: utf-8 -*-
import os
import tempfile

from marshmallow import Schema, fields

from ddb.feature.schema import FeatureSchema


class CookiecutterOptions(Schema):
    """
    Cookiecutter options
    """
    no_input = fields.Boolean(allow_none=True, default=None)
    replay = fields.Boolean(allow_none=True, default=False)
    overwrite_if_exists = fields.Boolean(allow_none=True, default=True)
    config_file = fields.String(allow_none=True, default=None)
    default_config = fields.Raw(allow_none=True, default=None)


class TemplateSchema(CookiecutterOptions):
    """
    Template schema
    """
    template = fields.String(required=True)
    output_dir = fields.String(allow_none=True, default=None)
    checkout = fields.String(allow_none=True, default=None)
    extra_context = fields.Dict(allow_none=True, default=None)
    password = fields.String(allow_none=True, default=None)
    version = fields.String(allow_none=True, default=None)


cookiecutter_tmp_dir = os.path.join(tempfile.gettempdir(), "ddb", "cookiecutter")


class CookiecutterFeatureSchema(FeatureSchema, CookiecutterOptions):
    """
    Cookiecutter feature schema.
    """
    templates = fields.List(fields.Nested(TemplateSchema()))
    cookiecutters_dir = fields.String(allow_none=True, default=os.path.join(cookiecutter_tmp_dir, "templates"))
    replay_dir = fields.String(allow_none=True, default=os.path.join(cookiecutter_tmp_dir, "replay"))
    default_context = fields.Dict(allow_none=True)



import os
import stat

import pytest

from ddb.__main__ import load_registered_features, register_actions_in_event_bus
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.jinja import JinjaFeature
from ddb.feature.file import FileFeature, FileWalkAction
from ddb.feature.permissions import PermissionsFeature


@pytest.mark.skipif("os.name == 'nt'")
class TestPermissionsAction:
    def test_project(self, project_loader):
        project_loader("project")
        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(PermissionsFeature())

        load_registered_features()
        register_actions_in_event_bus()

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.access("script.sh", os.X_OK)
        assert not os.access(os.path.join("subdirectory", "another-script.sh"), os.X_OK)

    def test_project2(self, project_loader):
        project_loader("project2")
        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(PermissionsFeature())

        load_registered_features()
        register_actions_in_event_bus()

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.access("script.sh", os.X_OK)
        assert os.access(os.path.join("subdirectory", "another-script.sh"), os.X_OK)

    def test_project3(self, project_loader):
        project_loader("project3")
        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(PermissionsFeature())

        load_registered_features()
        register_actions_in_event_bus()

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.access("script.sh", os.X_OK)
        assert os.access(os.path.join("subdirectory", "another-script.sh"), os.X_OK)
        assert os.access(os.path.join("subdirectory"), os.W_OK)
        assert not os.access(os.path.join("denied"), os.W_OK)

    def test_inherit_from_template(self, project_loader):
        project_loader("inherit_from_template")
        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        features.register(PermissionsFeature())

        load_registered_features()
        register_actions_in_event_bus()

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists("script.sh")
        assert os.access("script.sh", os.X_OK)



import os

import pytest
from pytest_mock import MockerFixture

from ddb.__main__ import main, load_registered_features
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.shell import ShellFeature
from ddb.feature.smartcd import SmartcdFeature, SmartcdAction, WindowsProjectActivate


@pytest.mark.skipif("os.name == 'nt'")
class TestSmartcdAction:
    def test_empty_project_without_core(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.smartcd.actions.is_smartcd_installed', lambda: True)

        project_loader("empty")

        features.register(SmartcdFeature())
        load_registered_features()

        action = SmartcdAction()
        action.execute()

        assert not os.path.exists(".bash_enter")
        assert not os.path.exists(".bash_leave")

    def test_empty_project_with_core(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.smartcd.actions.is_smartcd_installed', lambda: True)

        project_loader("empty")

        features.register(CoreFeature())
        features.register(SmartcdFeature())
        load_registered_features()

        action = SmartcdAction()
        action.execute()

        assert not os.path.exists(".bash_enter")
        assert not os.path.exists(".bash_leave")

    def test_empty_project_with_activate_deactivate_commands(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.smartcd.actions.is_smartcd_installed', lambda: True)

        project_loader("empty")

        features.register(CoreFeature())
        features.register(ShellFeature())
        features.register(SmartcdFeature())
        load_registered_features()

        action = SmartcdAction()
        action.execute()

        assert os.path.exists(".bash_enter")
        assert os.path.exists(".bash_leave")

        with open(".bash_enter") as f:
            content = f.read()
            assert "$(ddb activate)" in content

        with open(".bash_leave") as f:
            assert "$(ddb deactivate)" in f.read()

    def test_empty_project_main(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.smartcd.actions.is_smartcd_installed', lambda: True)

        project_loader("empty")

        main(["configure"])

        assert os.path.exists(".bash_enter")
        assert os.path.exists(".bash_leave")

        with open(".bash_enter") as f:
            content = f.read()
            assert "$(ddb activate)" in content

        with open(".bash_leave") as f:
            assert "$(ddb deactivate)" in f.read()

    def test_empty_project_main_no_smartcd(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.smartcd.actions.is_smartcd_installed', lambda: False)

        project_loader("empty")

        main(["configure"])

        assert not os.path.exists(".bash_enter")
        assert not os.path.exists(".bash_leave")


@pytest.mark.skipif("os.name != 'nt'")
class TestWindowsProjectActivate:
    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(SmartcdFeature())
        load_registered_features()

        action = WindowsProjectActivate()
        action.execute()

        assert not os.path.exists("ddb_activate.bat")
        assert not os.path.exists("ddb_deactivate.bat")

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(SmartcdFeature())
        load_registered_features()

        action = WindowsProjectActivate()
        action.execute()

        assert not os.path.exists("ddb_activate.bat")
        assert not os.path.exists("ddb_deactivate.bat")

    def test_empty_project_with_activate_deactivate_commands(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(ShellFeature())
        features.register(SmartcdFeature())
        load_registered_features()

        action = WindowsProjectActivate()
        action.execute()

        assert os.path.exists("ddb_activate.bat")
        assert os.path.exists("ddb_deactivate.bat")

        with open("ddb_activate.bat") as f:
            assert "set command=(ddb activate)" in f.read()

        with open("ddb_deactivate.bat") as f:
            assert "set command=(ddb deactivate)" in f.read()

    def test_empty_project_main(self, project_loader):
        project_loader("empty")

        main(["configure"])

        assert os.path.exists("ddb_activate.bat")
        assert os.path.exists("ddb_deactivate.bat")

        with open("ddb_activate.bat") as f:
            assert "set command=(ddb activate)" in f.read()

        with open("ddb_deactivate.bat") as f:
            assert "set command=(ddb deactivate)" in f.read()



# -*- coding: utf-8 -*-
"""
Checksums module
"""

import hashlib

from .crypto import convert_pem_to_der


def validate_checksums(response):
    """
    Validate checksums from certificate request response
    :param response:
    :return:
    """
    sums = response.get('sums', {})
    for part, content in response.items():
        if part not in ('sums', 'private_key'):
            sums_item = sums.get(part)
            if not sums_item:
                continue
            validate_checksum(part, content.encode('ascii'), sums_item)


def validate_checksum(part, content, sums_item, content_is_der=False):
    """
    Validate a checksum from certificate sums and content
    :param part:
    :param content:
    :param sums_item:
    :param content_is_der:
    :return:
    """
    for algo, checksum in sums_item.items():
        validate_checksum_algo(part, content, algo, checksum, content_is_der)


def validate_checksum_algo(part, content, algo, checksum, content_is_der=False):
    """
    Validate a checksum using content, algo and checksum.
    :param part:
    :param content:
    :param algo:
    :param checksum:
    :param content_is_der:
    :return:
    """
    if not content_is_der:
        content = convert_pem_to_der(part, content)

    validated_checksum = getattr(hashlib, algo.replace('-', ''))(content).hexdigest()
    checksum = checksum.lower()
    if validated_checksum != checksum:
        raise IOError("Invalid %s checksum (%s): %s != %s " % (part, algo, checksum, validated_checksum))



# -*- coding: utf-8 -*-
from typing import Iterable

from ddb.action import Action
from ddb.event import events
from ddb.feature import Feature


class Custom2Feature(Feature):
    """
    Custom Feature
    """

    @property
    def name(self) -> str:
        return "custom2"

    @property
    def actions(self) -> Iterable[Action]:
        return (
            Custom2Action(),
        )


class Custom2Action(Action):
    @property
    def event_bindings(self):
        return events.phase.configure

    @property
    def name(self):
        return "custom2_action"

    def execute(self):
        with open("test2", "w") as f:
            f.write("Custom2Action")



import os
import shlex
import subprocess
from typing import List

from ddb.config import config


def run(executable: str, *args: str):
    """
    Run executable using core.process configuration, replacing bin with configured one, appending and prepending args.
    """
    command_list = effective_command(executable, *args)
    process = subprocess.run(command_list,
                             check=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    if os.name == "nt":
        # On windows, there's ANSI code after output that has to be dropped...
        try:
            eof_index = process.stdout.index(b"\x1b[0m")
            process.stdout = process.stdout[:eof_index]
        except ValueError:
            pass
    return process.stdout


def effective_command(executable: str, *args: str) -> List[str]:
    """
    Get subprocess.run arguments using core.process configuration, replacing bin with configured one, appending and
    prepending args.
    """
    if not args:
        executable, *args = shlex.split(executable)

    configured_executable = config.data.get("core.process.%s.bin" % (executable,), executable)
    prepend_args = config.data.get("core.process.%s.prepend" % (executable,), [])
    append_args = config.data.get("core.process.%s.append" % (executable,), [])

    if os.name == 'nt' and not configured_executable.endswith('.exe'):
        configured_executable = "%s.exe" % (configured_executable,)

    if isinstance(prepend_args, str):
        prepend_args = shlex.split(prepend_args)

    if isinstance(append_args, str):
        append_args = shlex.split(append_args)

    return [configured_executable] + prepend_args + list(args) + append_args



import os
import pathlib
import re

import pytest
import yaml

from ddb.__main__ import load_registered_features, register_actions_in_event_bus
from ddb.config import config, migrations
from ddb.config.migrations import PropertyMigration
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.docker import DockerFeature
from ddb.feature.file import FileFeature, FileWalkAction
from ddb.feature.jsonnet import JsonnetFeature


class TestJsonnetAction:
    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(FileFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

    @pytest.mark.skipif("os.name == 'nt'")
    def test_named_user_group(self, project_loader):
        project_loader("empty")

        features.register(JsonnetFeature())
        load_registered_features()

        assert config.data.get('jsonnet.docker.user.name_to_uid')
        assert config.data.get('jsonnet.docker.user.group_to_gid')

    @pytest.mark.skipif("os.name != 'nt'")
    def test_named_user_group_windows(self, project_loader):
        project_loader("empty")

        features.register(JsonnetFeature())
        load_registered_features()

        assert config.data.get('jsonnet.docker.user.name_to_uid') == {}
        assert config.data.get('jsonnet.docker.user.group_to_gid') == {}

    def test_example1(self, project_loader):
        project_loader("example1")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('example1.json')
        with open('example1.json', 'r') as f:
            example = f.read()

        with open('example1.expected.json', 'r') as f:
            example_expected = f.read()

        assert example == example_expected

    def test_example1_yaml(self, project_loader):
        project_loader("example1.yaml")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('example1.another')
        with open('example1.another', 'r') as f:
            example_another = f.read()

        with open('example1.expected.another', 'r') as f:
            example_another_expected = f.read()

        assert example_another == example_another_expected

        assert os.path.exists('example1.yaml')
        with open('example1.yaml', 'r') as f:
            example_yaml = f.read()

        with open('example1.expected.yaml', 'r') as f:
            example_yaml_expected = f.read()

        assert example_yaml == example_yaml_expected

    def test_example2(self, project_loader):
        project_loader("example2")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('example2.json')
        with open('example2.json', 'r') as f:
            example = f.read()

        with open('example2.expected.json', 'r') as f:
            example_expected = f.read()

        assert example == example_expected

    def test_example3(self, project_loader):
        project_loader("example3")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('uwsgi.ini')
        with open('uwsgi.ini', 'r') as f:
            iwsgi = f.read()

        with open('uwsgi.expected.ini', 'r') as f:
            iwsgi_expected = f.read()

        assert iwsgi == iwsgi_expected

        assert os.path.exists('init.sh')
        with open('init.sh', 'r') as f:
            init = f.read()

        with open('init.expected.sh', 'r') as f:
            init_expected = f.read()

        assert init == init_expected

        assert os.path.exists('cassandra.conf')
        with open('cassandra.conf', 'r') as f:
            cassandra = f.read()

        with open('cassandra.expected.conf', 'r') as f:
            cassandra_expected = f.read()

        assert cassandra == cassandra_expected

    def test_example3_with_dir(self, project_loader):
        project_loader("example3.with_dir")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('./target/uwsgi.ini')
        with open('./target/uwsgi.ini', 'r') as f:
            iwsgi = f.read()

        with open('uwsgi.expected.ini', 'r') as f:
            iwsgi_expected = f.read()

        assert iwsgi == iwsgi_expected

        assert os.path.exists('./target/init.sh')
        with open('./target/init.sh', 'r') as f:
            init = f.read()

        with open('init.expected.sh', 'r') as f:
            init_expected = f.read()

        assert init == init_expected

        assert os.path.exists('./target/cassandra.conf')
        with open('./target/cassandra.conf', 'r') as f:
            cassandra = f.read()

        with open('cassandra.expected.conf', 'r') as f:
            cassandra_expected = f.read()

        assert cassandra == cassandra_expected

    def test_config_variables(self, project_loader):
        project_loader("config_variables")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('variables.json')
        with open('variables.json', 'r') as f:
            variables = f.read()

        with open('variables.expected.json', 'r') as f:
            variables_expected = f.read()

        assert variables == variables_expected

    @pytest.mark.parametrize("variant", [
        "test-dev",
        "test-ci",
        "test-stage",
        "test-prod",
    ])
    def test_docker_compose_traefik(self, project_loader, variant):
        def before_load_config():
            os.rename("ddb.%s.yml" % variant, "ddb.yml")
            os.rename("docker-compose.expected.%s.yml" % variant, "docker-compose.expected.yml")

        project_loader("docker_compose_traefik", before_load_config)

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()

            if os.name == 'nt':
                mapped_cwd = re.sub(r"^([a-zA-Z]):", r"/\1", os.getcwd())
                mapped_cwd = pathlib.Path(mapped_cwd).as_posix()

                expected_data = expected_data.replace("%ddb.path.project%", mapped_cwd)
            else:
                expected_data = expected_data.replace("%ddb.path.project%", os.getcwd())
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected_data = expected_data.replace("%uid%", str(config.data.get('docker.user.uid')))
            expected_data = expected_data.replace("%gid%", str(config.data.get('docker.user.gid')))
            expected_data = expected_data.replace("%docker.debug.host%", str(config.data.get('docker.debug.host')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected

    @pytest.mark.parametrize("variant", [
        "test-dev",
        "test-ci",
        "test-stage",
        "test-prod",
    ])
    def test_docker_compose_traefik_no_https(self, project_loader, variant):
        def before_load_config():
            os.rename("ddb.%s.yml" % variant, "ddb.yml")
            os.rename("docker-compose.expected.%s.yml" % variant, "docker-compose.expected.yml")

        project_loader("docker_compose_traefik_no_https", before_load_config)

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()

            if os.name == 'nt':
                mapped_cwd = re.sub(r"^([a-zA-Z]):", r"/\1", os.getcwd())
                mapped_cwd = pathlib.Path(mapped_cwd).as_posix()

                expected_data = expected_data.replace("%ddb.path.project%", mapped_cwd)
            else:
                expected_data = expected_data.replace("%ddb.path.project%", os.getcwd())
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected_data = expected_data.replace("%uid%", str(config.data.get('docker.user.uid')))
            expected_data = expected_data.replace("%gid%", str(config.data.get('docker.user.gid')))
            expected_data = expected_data.replace("%docker.debug.host%", str(config.data.get('docker.debug.host')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected

    @pytest.mark.parametrize("variant", [
        "dev",
        "ci",
        "prod",
    ])
    def test_docker_compose_traefik_defaults(self, project_loader, variant):
        def before_load_config():
            os.rename("ddb.%s.yml" % variant, "ddb.yml")
            os.rename("docker-compose.expected.%s.yml" % variant, "docker-compose.expected.yml")

        project_loader("docker_compose_traefik_defaults", before_load_config)

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()

            if os.name == 'nt':
                mapped_cwd = re.sub(r"^([a-zA-Z]):", r"/\1", os.getcwd())
                mapped_cwd = pathlib.Path(mapped_cwd).as_posix()

                expected_data = expected_data.replace("%ddb.path.project%", mapped_cwd)
            else:
                expected_data = expected_data.replace("%ddb.path.project%", os.getcwd())
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected_data = expected_data.replace("%uid%", str(config.data.get('docker.user.uid')))
            expected_data = expected_data.replace("%gid%", str(config.data.get('docker.user.gid')))
            expected_data = expected_data.replace("%docker.debug.host%", str(config.data.get('docker.debug.host')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected

    def test_docker_compose_variables(self, project_loader):
        project_loader("docker_compose_variables")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()

            if os.name == 'nt':
                mapped_cwd = re.sub(r"^([a-zA-Z]):", r"/\1", os.getcwd())
                mapped_cwd = pathlib.Path(mapped_cwd).as_posix()

                expected_data = expected_data.replace("%ddb.path.project%", mapped_cwd)
            else:
                expected_data = expected_data.replace("%ddb.path.project%", os.getcwd())
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected_data = expected_data.replace("%uid%", str(config.data.get('docker.user.uid')))
            expected_data = expected_data.replace("%gid%", str(config.data.get('docker.user.gid')))
            expected_data = expected_data.replace("%docker.debug.host%", str(config.data.get('docker.debug.host')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected

    def test_docker_compose_project_dot_com(self, project_loader):
        project_loader("docker_compose_project_dot_com")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()

            if os.name == 'nt':
                mapped_cwd = re.sub(r"^([a-zA-Z]):", r"/\1", os.getcwd())
                mapped_cwd = pathlib.Path(mapped_cwd).as_posix()

                expected_data = expected_data.replace("%ddb.path.project%", mapped_cwd)
            else:
                expected_data = expected_data.replace("%ddb.path.project%", os.getcwd())
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected_data = expected_data.replace("%uid%", str(config.data.get('docker.user.uid')))
            expected_data = expected_data.replace("%gid%", str(config.data.get('docker.user.gid')))
            expected_data = expected_data.replace("%docker.debug.host%", str(config.data.get('docker.debug.host')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected

    def test_docker_compose_excluded_services(self, project_loader):
        project_loader("docker_compose_excluded_services")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected

    def test_docker_compose_included_services(self, project_loader):
        project_loader("docker_compose_included_services")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected

    @pytest.mark.parametrize("variant", [
        "_register_binary",
        "_register_binary_with_one_option",
        # "_register_binary_with_multiple_options", TODO handle (options)(c1)
        "_shared_volumes",
        "_mount_volumes",
        "_mount_single_volume",
        "_mount_single_volume_with_default",
        "_expose"
    ])
    def test_docker_compose_variants(self, project_loader, variant):
        project_loader("docker_compose" + variant)

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()

            if os.name == 'nt':
                mapped_cwd = re.sub(r"^([a-zA-Z]):", r"/\1", os.getcwd())
                mapped_cwd = pathlib.Path(mapped_cwd).as_posix()

                expected_data = expected_data.replace("%ddb.path.project%", mapped_cwd)
            else:
                expected_data = expected_data.replace("%ddb.path.project%", os.getcwd())
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected_data = expected_data.replace("%uid%", str(config.data.get('docker.user.uid')))
            expected_data = expected_data.replace("%gid%", str(config.data.get('docker.user.gid')))
            expected_data = expected_data.replace("%docker.debug.host%", str(config.data.get('docker.debug.host')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected

        if variant == '_mount_single_volume':
            assert os.path.isdir('volumes/shared-volume')

        if variant == '_mount_single_volume_with_default':
            assert os.path.isdir('shared-volume')

    @pytest.mark.parametrize("variant", [
        "default",
        "no_debug",
    ])
    def test_docker_compose_xdebug(self, project_loader, variant):
        def before_load_config():
            os.rename("ddb.%s.yml" % variant, "ddb.yml")
            os.rename("docker-compose.expected.%s.yml" % variant, "docker-compose.expected.yml")

        project_loader("docker_compose_xdebug", before_load_config)

        print(os.getcwd())

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            rendered = yaml.load(f.read(), yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()

            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected_data = expected_data.replace("%uid%", str(config.data.get('docker.user.uid')))
            expected_data = expected_data.replace("%gid%", str(config.data.get('docker.user.gid')))
            expected_data = expected_data.replace("%docker.debug.host%", str(config.data.get('docker.debug.host')))

            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert rendered == expected


class TestJsonnetAutofix:
    def teardown_method(self, test_method):
        migrations.set_history()

    def test_autofix_variables_only(self, project_loader):
        project_loader("autofix_variables_only")

        config.args.autofix = True

        history = (
            PropertyMigration("old_property",
                              "new_property", since="v1.1.0"),
            PropertyMigration("some.deep.old.property",
                              "some.another.new.property", since="v1.1.0"),
        )

        migrations.set_history(history)

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('variables.json')
        with open('variables.json', 'r') as f:
            rendered = f.read()

        with open('variables.expected.json', 'r') as f:
            expected = f.read()

        assert expected == rendered

        with open('variables.json.jsonnet', 'r') as f:
            source = f.read()

        with open('variables.json.autofix', 'r') as f:
            fixed = f.read()

        assert source == fixed



import logging
from typing import Optional, List, Callable, Dict
from typing import TYPE_CHECKING

from dotty_dict import Dotty

if TYPE_CHECKING:
    from ..action import Action  # pylint:disable=cyclic-import


class ContextStackItem:
    """
    A stack item inside context
    """

    def __init__(self, event_name: str, action: 'Action', to_call: Callable, args: list, kwargs: dict):
        self.event_name = event_name
        self.action = action
        self.to_call = to_call
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def _build_parameters_repr(*args, **kwargs):
        parameters_repr = ', '.join(map(str, args))
        for key, value in sorted(kwargs.items()):
            if parameters_repr:
                parameters_repr += ', '
            parameters_repr += key + "=" + str(value)
        return parameters_repr

    def __repr__(self):
        parameters_repr = self._build_parameters_repr(*self.args, **self.kwargs)
        return "%s => %s.%s(%s)" % (self.event_name,
                                    type(self.action).__name__,
                                    self.to_call.__name__,
                                    parameters_repr)


class Context:  # pylint:disable=too-many-instance-attributes
    """
    Execution context.
    """

    def __init__(self):
        self.phase = None  # type: Optional['Phase']
        self.command = None  # type: Optional['Command']
        self.stack = []  # type: List[ContextStackItem]
        self.watching = False  # type: bool
        self.exceptions = []  # type: List[Exception]
        self.processed_sources = {}  # type: Dict[str, str]
        self.processed_targets = {}  # type: Dict[str, str]
        self.data = Dotty({})

    def reset(self):
        """
        Reset the context.
        """
        self.__init__()

    @property
    def event_name(self) -> str:
        """
        Current event name in stack.
        """
        return self.stack[-1].event_name if self.stack else None

    @property
    def action(self) -> 'Action':
        """
        Current action in stack.
        """
        return self.stack[-1].action if self.stack else None

    @property
    def to_call(self) -> Callable:
        """
        Current callable in stack
        """
        return self.stack[-1].to_call if self.stack else None

    @property
    def log(self):
        """
        Logger for current context.
        """
        logger_name = ".".join(["ddb.context"] +
                               list(map(lambda x: x.name,
                                        [x for x in [self.command, self.phase, self.action] if x]))
                               )
        return logging.getLogger(logger_name)

    @property
    def logger(self):
        """
        Logger for current context.
        """
        return self.log

    def mark_as_processed(self, source, target):
        """
        Mark sources and target as processed, for them to be skipped by other file template finders.
        """
        self.processed_sources[source] = target
        self.processed_targets[target] = source

    def mark_as_unprocessed(self, filepath=None):
        """
        Mark sources and target as processed, for them to be skipped by other file template finders.
        """
        if filepath and filepath in self.processed_sources:
            linked_target = self.processed_sources.pop(filepath)
            if linked_target in self.processed_targets:
                self.processed_targets.pop(linked_target)

        if filepath in self.processed_targets:
            linked_source = self.processed_targets.pop(filepath)
            if linked_source in self.processed_sources:
                self.processed_sources.pop(linked_source)



import os

import pytest
import yaml

from ddb.__main__ import main
from ddb.config import config
from tests.utilstest import expect_gitignore, setup_cfssl


class TestEject:
    @pytest.mark.docker
    def test_eject1(self, project_loader, module_scoped_container_getter):
        project_loader("eject1")

        setup_cfssl(module_scoped_container_getter)

        main(["configure"])

        assert os.path.exists("docker-compose.yml")
        assert os.path.exists("docker-compose.yml.jsonnet")
        assert expect_gitignore(".gitignore", "/docker-compose.yml")

        assert os.path.exists(os.path.join(".bin", "psql" + (".bat" if os.name == "nt" else "")))
        assert expect_gitignore(".gitignore", "/.bin/psql" + (".bat" if os.name == "nt" else ""))

        assert os.path.exists(os.path.join(".docker", "db", "Dockerfile.jinja"))
        assert expect_gitignore(".gitignore", "/.docker/db/Dockerfile")

        main(["configure", "--eject"], reset_disabled=True)

        assert os.path.exists("docker-compose.yml")
        assert not os.path.exists("docker-compose.yml.jsonnet")
        assert not expect_gitignore(".gitignore", "/docker-compose.yml")

        assert os.path.exists(os.path.join(".bin", "psql" + (".bat" if os.name == "nt" else "")))
        assert expect_gitignore(".gitignore", "/.bin/psql" + (".bat" if os.name == "nt" else ""))

        assert not os.path.exists(os.path.join(".docker", "db", "Dockerfile.jinja"))
        assert not expect_gitignore(".gitignore", "/.docker/db/Dockerfile")

        with open('docker-compose.yml', 'r') as dc_file:
            actual = yaml.load(dc_file, yaml.SafeLoader)

        with open('../expected/docker-compose.yml', 'r') as expected_dc_file:
            expected_data = expected_dc_file.read()
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert actual == expected

    @pytest.mark.docker
    def test_eject2(self, project_loader, module_scoped_container_getter):
        project_loader("eject2")

        setup_cfssl(module_scoped_container_getter)

        main(["configure"])

        assert os.path.exists("docker-compose.yml")
        assert os.path.exists("docker-compose.yml.jsonnet")
        assert expect_gitignore(".gitignore", "/docker-compose.yml")

        assert os.path.exists(os.path.join(".bin", "psql" + (".bat" if os.name == "nt" else "")))
        assert expect_gitignore(".gitignore", "/.bin/psql" + (".bat" if os.name == "nt" else ""))

        assert os.path.exists(os.path.join(".docker", "db", "Dockerfile.jinja"))
        assert expect_gitignore(".gitignore", "/.docker/db/Dockerfile")

        main(["configure", "--eject"], reset_disabled=True)

        assert os.path.exists("docker-compose.yml")
        assert not os.path.exists("docker-compose.yml.jsonnet")
        assert not expect_gitignore(".gitignore", "/docker-compose.yml")

        assert os.path.exists(os.path.join(".bin", "psql" + (".bat" if os.name == "nt" else "")))
        assert expect_gitignore(".gitignore", "/.bin/psql" + (".bat" if os.name == "nt" else ""))

        assert not os.path.exists(os.path.join(".docker", "db", "Dockerfile.jinja"))
        assert not expect_gitignore(".gitignore", "/.docker/db/Dockerfile")

        with open('docker-compose.yml', 'r') as dc_file:
            actual = yaml.load(dc_file, yaml.SafeLoader)

        with open('../expected/docker-compose.yml', 'r') as expected_dc_file:
            expected_data = expected_dc_file.read()
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert actual == expected

    @pytest.mark.docker
    def test_eject2_with_jsonnet_disabled(self, project_loader, module_scoped_container_getter):
        project_loader("eject2")

        setup_cfssl(module_scoped_container_getter)

        main(["configure"])

        assert os.path.exists("docker-compose.yml")
        assert os.path.exists("docker-compose.yml.jsonnet")
        assert expect_gitignore(".gitignore", "/docker-compose.yml")

        assert os.path.exists(os.path.join(".bin", "psql" + (".bat" if os.name == "nt" else "")))
        assert expect_gitignore(".gitignore", "/.bin/psql" + (".bat" if os.name == "nt" else ""))

        assert os.path.exists(os.path.join(".docker", "db", "Dockerfile.jinja"))
        assert expect_gitignore(".gitignore", "/.docker/db/Dockerfile")

        os.environ['DDB_OVERRIDE_JSONNET_DOCKER_VIRTUALHOST_DISABLED'] = "1"
        os.environ['DDB_OVERRIDE_JSONNET_DOCKER_BINARY_DISABLED'] = "True"
        main(["configure", "--eject"], reset_disabled=True)

        assert os.path.exists("docker-compose.yml")
        assert not os.path.exists("docker-compose.yml.jsonnet")
        assert not expect_gitignore(".gitignore", "/docker-compose.yml")

        assert not os.path.exists(os.path.join(".bin", "psql" + (".bat" if os.name == "nt" else "")))
        assert not expect_gitignore(".gitignore", "/.bin/psql" + (".bat" if os.name == "nt" else ""))

        assert not os.path.exists(os.path.join(".docker", "db", "Dockerfile.jinja"))
        assert not expect_gitignore(".gitignore", "/.docker/db/Dockerfile")

        with open('docker-compose.yml', 'r') as dc_file:
            actual = yaml.load(dc_file, yaml.SafeLoader)

        with open('../expected/docker-compose.jsonnet.disabled.yml', 'r') as expected_dc_file:
            expected_data = expected_dc_file.read()
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            expected = yaml.load(expected_data, yaml.SafeLoader)

        assert actual == expected

    def test_eject3(self, project_loader):
        project_loader("eject3")

        main(["configure"])

        assert os.path.exists("../home/test.txt")
        assert os.path.exists(".docker/c1/test.txt")
        assert os.path.exists(".docker/c2/test.txt")
        assert os.path.exists("config.jinja.txt")
        assert os.path.exists("config.txt")

        main(["configure", "--eject"])

        assert os.path.exists("../home/test.txt")
        assert os.path.exists(".docker/c1/test.txt")
        assert os.path.exists(".docker/c2/test.txt")
        assert not os.path.exists("config.jinja.txt")
        assert os.path.exists("config.txt")



# -*- coding: utf-8 -*-
import os
from argparse import Namespace
from collections import namedtuple
from copy import deepcopy
from os.path import exists
from pathlib import Path
from typing import Union, Iterable, Dict

import yaml

from ddb.config.merger import config_merger
from ddb.config.migrations import MigrationsDotty

ConfigPaths = namedtuple('ConfigPaths', ['ddb_home', 'home', 'project_home'])


def configuration_file(path: str, filenames: Iterable[str], extensions: Iterable[str]):
    """
    Find configuration file for given path and possible filename/extensions
    """
    if path:
        for basename in filenames:
            for ext in extensions:
                file = os.path.join(path, basename + '.' + ext)
                if exists(file):
                    return file
    return None


def get_default_config_paths(env_prefix, filenames, extensions) -> ConfigPaths:
    """
    Get configuration paths
    """
    if os.environ.get(env_prefix + '_PROJECT_HOME'):
        project_home = os.environ.get(env_prefix + '_PROJECT_HOME')
    else:
        project_home_candidate = os.getcwd()
        while not configuration_file(project_home_candidate, filenames, extensions):
            project_home_candidate_parent = str(Path(project_home_candidate).parent)
            if project_home_candidate_parent == project_home_candidate:
                project_home_candidate = os.getcwd()
                break
            project_home_candidate = project_home_candidate_parent
        project_home = project_home_candidate

    project_home = os.path.abspath(project_home)

    home = os.environ.get(env_prefix + '_HOME', os.path.join(str(Path.home()), '.docker-devbox'))
    ddb_home = os.environ.get(env_prefix + '_DDB_HOME', os.path.join(home, 'ddb'))

    return ConfigPaths(ddb_home=ddb_home, home=home, project_home=project_home)


class Config:  # pylint:disable=too-many-instance-attributes
    """
    Configuration
    """
    # Static default/overrides values for configuration, can be modified from code mainly for tests purpose
    defaults = None
    overrides = lambda config: config

    def __init__(self,
                 paths: Union[ConfigPaths, None] = None,
                 env_prefix='DDB',
                 env_override_prefix='DDB_OVERRIDE',
                 filenames=('ddb', 'ddb.local'),
                 extensions=('yml', 'yaml'),
                 cwd=None,
                 args=None,
                 unknown_args=None):
        self.env_prefix = env_prefix
        self.env_override_prefix = env_override_prefix
        self.filenames = filenames
        self.extensions = extensions
        self.env_additions = {}
        self.data = MigrationsDotty()
        self.paths = paths if paths else get_default_config_paths(env_prefix, filenames, extensions)
        self.cwd = cwd
        self.args = args if args else Namespace()
        self.unknown_args = unknown_args if unknown_args else []
        self.env_key = 'env'

    def reset(self, *args, **kwargs):
        """
        Reset the configuration object, while keeping configured paths.
        """
        self.__init__(*args, paths=self.paths, **kwargs)

    def clear(self):
        """
        Remove all configuration data.
        """
        self.data.clear()
        self.env_additions.clear()

    @property
    def project_configuration_file(self):
        """
        Retrieve the project configuration file.
        """
        return configuration_file(self.paths.project_home, self.filenames, self.extensions)

    @property
    def project_cwd(self):
        """
        Retrieve the current working directory, relative to project_home.
        :return:
        """
        if self.paths.project_home and self.cwd:
            return os.path.relpath(self.cwd, self.paths.project_home)
        return None

    @property
    def eject(self):
        """
        Check if command line is running configure command with --eject flag.
        :return:
        """
        return 'command' in self.args and self.args.command == 'configure' and \
               'eject' in self.args and self.args.eject

    @property
    def clear_cache(self):
        """
        Should cache be cleared on load
        """
        if 'clear_cache' in self.args:
            return self.args.clear_cache
        return False

    @property
    def files(self):
        """
        Possible configuration files to load.
        """
        ret = []
        for path in self.paths:
            if not path:
                continue
            for basename in self.filenames:
                for ext in self.extensions:
                    if basename.endswith('.' + ext) and os.path.exists(os.path.join(path, basename)):
                        file = os.path.join(path, basename)
                    else:
                        file = os.path.join(path, basename + '.' + ext)
                    ret.append(file)
        return ret

    def read(self, defaults=None, files=None):
        """
        Read configuration data from files.
        """
        if defaults is None:
            loaded_data = {}
        else:
            loaded_data = dict(defaults)

        if files is None:
            files = self.files

        found_files = {}

        for file in files:
            if exists(file):
                with open(file, 'rb') as stream:
                    file_data = yaml.safe_load(stream)
                    found_files[file] = self.apply_environ_overrides(deepcopy(file_data))
                    if file_data:
                        loaded_data = config_merger.merge(loaded_data, file_data)

        if Config.overrides:  # pylint:disable=using-constant-test
            Config.overrides(loaded_data)

        return self.apply_environ_overrides(loaded_data), found_files

    def load_from_data(self, data: Dict):
        """
        Load configuration data from readen files.
        """
        merged_data = config_merger.merge(dict(self.data.raw()), data)
        self.data = MigrationsDotty(merged_data)

        if self.env_key in self.data:
            env = self.data.pop(self.env_key)
            if env:
                for (name, value) in env.items():
                    os.environ[name] = value

    def load(self, defaults=None, files=None):
        """
        Load configuration data from files. Variable in 'env_key' key will be placed loaded as environment variables.
        """
        defaults_to_apply = {} if Config.defaults is None else dict(Config.defaults)
        if defaults:
            # Config.defaults should have the priority over given defaults.
            defaults_to_apply = config_merger.merge(defaults, defaults_to_apply)
        read_data, _ = self.read(defaults_to_apply, files)
        self.load_from_data(read_data)

    def apply_environ_overrides(self, data, prefix=None):
        """
        Apply environment variables to configuration.
        """
        if not prefix:
            prefix = self.env_override_prefix
        prefix = prefix.upper()

        environ_value = os.environ.get(prefix)
        if environ_value:
            if environ_value.lower() == str(True):
                environ_value = True
            elif environ_value.lower() == str(False):
                environ_value = False
            elif environ_value.isdigit():
                environ_value = int(environ_value)
            if isinstance(data, bool):
                environ_value = bool(environ_value)
            return environ_value

        if isinstance(data, dict):
            for (name, value) in data.items():
                key_prefix = prefix + "_" + name
                key_prefix = key_prefix.upper()

                data[name] = self.apply_environ_overrides(value, key_prefix)
        if isinstance(data, list):
            i = 0
            for value in data:
                replace_prefix = prefix + "[" + str(i) + "]"
                replace_prefix = replace_prefix.upper()

                environ_value = os.environ.get(replace_prefix)
                if environ_value:
                    data[i] = self.apply_environ_overrides(value, replace_prefix)

                i += 1

        return data



import os

from ddb.__main__ import load_registered_features, register_default_caches
from ddb.feature import features
from ddb.feature.copy import CopyFeature, CopyAction


class TestCopyFeature:
    def test_copy_ca_certificates(self, project_loader):
        project_loader("copy-ca-certificates")

        features.register(CopyFeature())
        load_registered_features()

        action = CopyAction()
        action.execute()

        assert os.path.exists(os.path.join('.docker', 'service1', 'ca-certs', 'some-cert.crt'))
        assert not os.path.exists(os.path.join('.docker', 'service1', 'ca-certs', 'another-cert.crt'))

        assert os.path.exists(os.path.join('.docker', 'service2', 'ca-certs', 'some-cert.crt'))
        assert not os.path.exists(os.path.join('.docker', 'service2', 'ca-certs', 'another-cert.crt'))

        assert not os.path.exists(os.path.join('.docker', '.not-a-service', 'ca-certs', 'some-cert.crt'))
        assert not os.path.exists(os.path.join('.docker', '.not-a-service', 'ca-certs', 'another-cert.crt'))

    def test_copy_fixuid(self, project_loader):
        project_loader("copy-fixuid")

        features.register(CopyFeature())
        load_registered_features()
        register_default_caches()

        action = CopyAction()
        action.execute()

        assert os.path.exists(os.path.join('.docker', 'service1', 'fixuid.tar.gz'))
        assert os.path.exists(os.path.join('.docker', 'service2', 'fixuid.tar.gz'))
        assert not os.path.exists(os.path.join('.docker', '.not-a-service', 'fixuid.tar.gz'))

    def test_copy_fixuid_default_filename(self, project_loader):
        project_loader("copy-fixuid-default-filename")

        features.register(CopyFeature())
        load_registered_features()
        register_default_caches()

        action = CopyAction()
        action.execute()

        assert os.path.exists(os.path.join('.docker', 'service1', 'fixuid-0.5-linux-amd64.tar.gz'))
        assert os.path.exists(os.path.join('.docker', 'service2', 'fixuid-0.5-linux-amd64.tar.gz'))
        assert not os.path.exists(os.path.join('.docker', '.not-a-service', 'fixuid-0.5-linux-amd64.tar.gz'))

    def test_copy_no_dispatch(self, project_loader):
        project_loader("copy-no-dispatch")

        features.register(CopyFeature())
        load_registered_features()
        register_default_caches()

        action = CopyAction()
        action.execute()
        assert os.path.exists(os.path.join('.', 'some-cert.crt'))

    def test_copy_subdirectory(self, project_loader):
        project_loader("copy-subdirectory")

        features.register(CopyFeature())
        load_registered_features()
        register_default_caches()

        action = CopyAction()
        action.execute()
        assert os.path.exists(os.path.join('sub', 'target'))

        assert os.path.isfile(os.path.join('sub', 'target', 'test.abc.json'))
        assert os.path.isfile(os.path.join('sub', 'target', 'test.def.json'))
        assert os.path.isfile(os.path.join('sub', 'target', 'test.ijk.yaml'))

    def test_copy_subdirectory_glob(self, project_loader):
        project_loader("copy-subdirectory-glob")

        features.register(CopyFeature())
        load_registered_features()
        register_default_caches()

        action = CopyAction()
        action.execute()
        assert os.path.exists(os.path.join('sub', 'target'))

        assert os.path.isfile(os.path.join('sub', 'target', 'test.abc.json'))
        assert os.path.isfile(os.path.join('sub', 'target', 'test.def.json'))
        assert not os.path.exists(os.path.join('sub', 'target', 'test.ijk.yaml'))



from deepmerge import Merger
from deepmerge.strategy.dict import DictStrategies
from deepmerge.strategy.fallback import FallbackStrategies
from deepmerge.strategy.list import ListStrategies
from deepmerge.strategy.type_conflict import TypeConflictStrategies


def _type_conflict_strategy(config, path, base, nxt):
    strategy = TypeConflictStrategies.strategy_override
    if isinstance(nxt, dict):
        if 'merge' in nxt and 'value' in nxt:
            value = nxt['value']
            merge = nxt['merge']
            if isinstance(merge, str):
                if isinstance(value, list) and isinstance(base, list):
                    strategy = ExtendedListStrategies._expand_strategy(merge)  # pylint:disable=protected-access
                    nxt = value
                elif isinstance(value, dict) and isinstance(base, dict):
                    strategy = DictStrategies._expand_strategy(merge)  # pylint:disable=protected-access
                    nxt = value
    return strategy(config, path, base, nxt)


def clean_merge_value_dicts(result):
    """

    :return:
    """

    if isinstance(result, dict):
        updates = []
        for key, value in result.items():
            if isinstance(value, dict) and 'merge' in value and 'value' in value:
                updates.append((key, value['value']))
            elif isinstance(value, list):
                clean_merge_value_dicts(value)
        for key, value in updates:
            result[key] = value

    if isinstance(result, list):
        for elt in result:
            clean_merge_value_dicts(elt)


class ExtendedListStrategies(ListStrategies):
    """
    Custom list strategies
    """

    @staticmethod
    def strategy_append_if_missing(config, path, base, nxt):
        """ use the list nxt. """
        filtered_nxt = [item for item in nxt if item not in base]
        return ListStrategies.strategy_append(config, path, base, filtered_nxt)

    @staticmethod
    def strategy_prepend_if_missing(config, path, base, nxt):
        """ use the list nxt. """
        filtered_nxt = [item for item in nxt if item not in base]
        return ListStrategies.strategy_prepend(config, path, base, filtered_nxt)

    @staticmethod
    def strategy_insert_if_missing(config, path, base, nxt):
        """ use the list nxt. """
        idx = nxt.pop(0)
        filtered_nxt = [item for item in nxt if item not in base]
        if idx >= 0:
            filtered_nxt = list(reversed(filtered_nxt))
        ret = list(base)
        for filtered_nxt_item in filtered_nxt:
            ret.insert(idx, filtered_nxt_item)
        return ret

    @staticmethod
    def strategy_insert(config, path, base, nxt):
        """ use the list nxt. """
        idx = nxt.pop(0)
        ordered_nxt = list(nxt)
        if idx >= 0:
            ordered_nxt = list(reversed(ordered_nxt))
        ret = list(base)
        for ordered_nxt_item in ordered_nxt:
            ret.insert(idx, ordered_nxt_item)
        return ret


def merge_value_strategy_wrapper(strategy):
    """
    Wrap the strategy to remove merge/value dicts if output still have ones.
    """

    def wrapped_strategy(config, path, base, nxt):
        result = strategy(config, path, base, nxt)
        clean_merge_value_dicts(result)
        return result

    return wrapped_strategy


config_merger = Merger(
    [
        (list, merge_value_strategy_wrapper(ListStrategies.strategy_override)),
        (dict, merge_value_strategy_wrapper(DictStrategies.strategy_merge))
    ],
    [merge_value_strategy_wrapper(FallbackStrategies.strategy_override)],
    [_type_conflict_strategy]
)



import os
import shutil
import threading
import time
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Iterable

import pytest
from ddb.__main__ import main, wait_watch_started, stop_watch
from tests.utilstest import expect_gitignore, setup_cfssl
from waiting import wait


def init_test_watch(watch: bool, command: Iterable[str]):
    thread = None

    full_command = ["--watch"] if watch else []
    full_command.extend(command)

    main_runner = lambda: main(full_command)

    if watch:
        thread = threading.Thread(name="watch",
                                  target=main_runner)
        thread.start()
        wait_watch_started(30)
        if os.environ.get("TRAVIS") == "true":
            # On TravisCI, it seems there is some race condition that may cause tests to fail without this.
            time.sleep(1)
    else:
        main_runner()

    return thread, main_runner


class TestWatch:
    @pytest.mark.parametrize("watch", [
        True,
        False
    ])
    def test_watch_file_change(self, project_loader, watch):
        project_loader("watch1")

        thread, main_runner = init_test_watch(watch, ["configure"])

        try:
            assert os.path.exists("test.txt")

            with open("test.txt.jinja", "w") as template_file:
                template_file.write("This is {{core.project.name}} project. (modified)")

            if not watch:
                main_runner()

            wait(lambda: os.path.exists("test.txt") and
                         Path("test.txt").read_text() == "This is watch1 project. (modified)",
                 timeout_seconds=5)

            wait(lambda: os.path.exists(".gitignore") and expect_gitignore(".gitignore", "/test.txt"),
                 timeout_seconds=5)
        finally:
            if watch:
                stop_watch()

        if thread:
            thread.join()

    @pytest.mark.parametrize("watch", [
        True,
        False
    ])
    def test_watch_file_created(self, project_loader, watch):
        project_loader("watch1")

        thread, main_runner = init_test_watch(watch, ["configure"])

        try:
            assert os.path.exists("test.txt")

            with open("test.created.txt.jinja", "w") as template_file:
                template_file.write("This is {{core.project.name}} project. (created)")

            if not watch:
                main_runner()

            wait(lambda: os.path.exists("test.created.txt") and
                         Path("test.created.txt").read_text() == "This is watch1 project. (created)",
                 timeout_seconds=5)

            wait(lambda: os.path.exists(".gitignore") and
                         expect_gitignore(".gitignore", "/test.txt", "/test.created.txt"),
                 timeout_seconds=5)
        finally:
            if watch:
                stop_watch()

        if thread:
            thread.join()

    @pytest.mark.parametrize("watch", [
        True,
        False
    ])
    def test_watch_file_delete(self, project_loader, watch):
        project_loader("watch1")

        thread, main_runner = init_test_watch(watch, ["configure"])

        try:
            assert os.path.exists("test.txt")

            os.remove("test.txt.jinja")

            if not watch:
                main_runner()

            wait(lambda: not os.path.exists("test.txt"), timeout_seconds=5)
            wait(lambda: not expect_gitignore(".gitignore", "/text.txt"), timeout_seconds=5)
        finally:
            if watch:
                stop_watch()

        if thread:
            thread.join()

    @pytest.mark.parametrize("watch", [
        True,
        False
    ])
    def test_watch_file_move(self, project_loader, watch):
        project_loader("watch1")

        thread, main_runner = init_test_watch(watch, ["configure"])

        try:
            assert os.path.exists("test.txt")

            os.rename("test.txt.jinja", "test2.txt.jinja")

            if not watch:
                main_runner()

            wait(lambda: not os.path.exists("test.txt") and os.path.exists("test2.txt") and "This is watch1 project.",
                 timeout_seconds=5)

            wait(lambda: os.path.exists(".gitignore") and expect_gitignore(".gitignore", "/test2.txt"),
                 timeout_seconds=5)
        finally:
            if watch:
                stop_watch()

        if thread:
            thread.join()


def copy_from_files(f):
    with NamedTemporaryFile('w', delete=False) as tmp:
        with open(os.path.join("..", "files", f)) as source:
            shutil.copyfileobj(source, tmp)

    os.rename(tmp.name, f)


class TestWatchFixuid:
    @pytest.mark.parametrize("watch", [
        True,
        False
    ])
    def test_watch_fixuid_order_1(self, project_loader, watch):
        project_loader("watch-fixuid")

        thread, main_runner = init_test_watch(watch, ["configure"])

        try:
            os.makedirs(os.path.join(".docker", "db"), exist_ok=True)

            copy_from_files(os.path.join(".docker", "db", "Dockerfile.jinja"))
            assert os.path.exists(os.path.join(".docker", "db", "Dockerfile.jinja"))

            copy_from_files(os.path.join(".docker", "db", "fixuid.yml"))
            assert os.path.exists(os.path.join(".docker", "db", "fixuid.yml"))

            copy_from_files("docker-compose.yml.jsonnet")
            assert os.path.exists("docker-compose.yml.jsonnet")

            if not watch:
                main_runner()

            wait(lambda: os.path.exists(os.path.join(".docker", "db", "Dockerfile")),
                 timeout_seconds=30)

            wait(lambda: os.path.exists("docker-compose.yml"),
                 timeout_seconds=30)

            def check_dockerfile():
                with open(os.path.join("..", "expected", "Dockerfile"), "r") as file:
                    expected_dockerfile = file.read()

                with open(os.path.join(".docker", "db", "Dockerfile"), "r") as file:
                    actual_dockerfile = file.read()

                return expected_dockerfile == actual_dockerfile

            wait(check_dockerfile, timeout_seconds=30)

        finally:
            if watch:
                stop_watch()

        if thread:
            thread.join()

    @pytest.mark.parametrize("parameters", [
        [True, ["docker-compose.yml.jsonnet",
                os.path.join(".docker", "db", "Dockerfile.jinja"),
                os.path.join(".docker", "db", "fixuid.yml")]],
        [False, ["docker-compose.yml.jsonnet",
                 os.path.join(".docker", "db", "Dockerfile.jinja"),
                 os.path.join(".docker", "db", "fixuid.yml")]],
        [True, [os.path.join(".docker", "db", "Dockerfile.jinja"),
                os.path.join(".docker", "db", "fixuid.yml"),
                "docker-compose.yml.jsonnet", ]],
        [False, [os.path.join(".docker", "db", "Dockerfile.jinja"),
                 os.path.join(".docker", "db", "fixuid.yml"),
                 "docker-compose.yml.jsonnet"]],
        [True, [os.path.join(".docker", "db", "fixuid.yml"),
                "docker-compose.yml.jsonnet",
                os.path.join(".docker", "db", "Dockerfile.jinja")]],
        [False, [os.path.join(".docker", "db", "fixuid.yml"),
                 "docker-compose.yml.jsonnet",
                 os.path.join(".docker", "db", "Dockerfile.jinja")]],
    ])
    def test_watch_fixuid_various_order(self, project_loader, parameters):
        watch, files = parameters
        project_loader("watch-fixuid")

        thread, main_runner = init_test_watch(watch, ["configure"])

        try:
            os.makedirs(os.path.join(".docker", "db"), exist_ok=True)

            for f in files:
                copy_from_files(f)
                time.sleep(0.5)

            if not watch:
                main_runner()

            wait(lambda: os.path.exists(os.path.join(".docker", "db", "Dockerfile")),
                 timeout_seconds=30)

            wait(lambda: os.path.exists("docker-compose.yml"),
                 timeout_seconds=30)

            wait(lambda: os.path.exists(os.path.join(".docker", "db", "fixuid.tar.gz")),
                 timeout_seconds=30)

            def _check_dockerfile():
                with open(os.path.join("..", "expected", "Dockerfile"), "r") as file:
                    expected_dockerfile = file.read()

                with open(os.path.join(".docker", "db", "Dockerfile"), "r") as file:
                    actual_dockerfile = file.read()

                return expected_dockerfile == actual_dockerfile

            wait(_check_dockerfile, timeout_seconds=30)

        finally:
            if watch:
                stop_watch()

        if thread:
            thread.join()


class TestWatchConfig:
    def test_watch_config_change(self, project_loader):
        project_loader("watch-config")

        thread, main_runner = init_test_watch(True, ["configure"])

        try:
            wait(lambda: os.path.exists("test.txt") and Path("test.txt").read_text() == "app.value: foo",
                 timeout_seconds=60)

            with open("ddb.yml", "r") as f:
                ddb_yml_content = f.read()

            ddb_yml_content = ddb_yml_content.replace("foo", "bar")
            with open("ddb.yml", "w") as f:
                f.write(ddb_yml_content)

            time.sleep(5)

            wait(lambda: os.path.exists("test.txt") and Path("test.txt").read_text() == "app.value: bar",
                 timeout_seconds=60)
        finally:
            stop_watch()

        if thread:
            thread.join()


class TestWatchCerts:
    @pytest.mark.parametrize("watch", [
        True,
        False
    ])
    @pytest.mark.docker
    def test_watch_certificates_when_virtualhost_is_added(self, watch, project_loader, module_scoped_container_getter):
        project_loader("watch-certs")

        setup_cfssl(module_scoped_container_getter)

        thread, main_runner = init_test_watch(watch, ["configure"])

        try:
            assert not os.path.exists(os.path.join(".certs", "watch-certs.test.key"))
            assert not os.path.exists(os.path.join(".certs", "watch-certs.test.crt"))

            copy_from_files("docker-compose.yml.jsonnet")

            if not watch:
                main_runner()

            wait(lambda: os.path.exists("docker-compose.yml"), timeout_seconds=5)
            wait(lambda: not os.path.exists(os.path.join(".certs", "watch-certs.test.key")), timeout_seconds=5)
            wait(lambda: not os.path.exists(os.path.join(".certs", "watch-certs.test.crt")), timeout_seconds=5)

            with open("docker-compose.yml.jsonnet", "r") as f:
                ddb_yml_content = f.read()

            ddb_yml_content = ddb_yml_content.replace('ddb.Image("httpd")',
                                                      'ddb.Image("httpd") + ddb.VirtualHost("80", "watch-certs.test", "httpd")')
            with open("docker-compose.yml.jsonnet", "w") as f:
                f.write(ddb_yml_content)

            if not watch:
                main_runner()

            wait(lambda: os.path.exists(os.path.join(".certs", "watch-certs.test.key")), timeout_seconds=5)
            wait(lambda: os.path.exists(os.path.join(".certs", "watch-certs.test.crt")), timeout_seconds=5)
            wait(lambda: os.path.exists(os.path.join("..", "home", "certs", "watch-certs.test.key")), timeout_seconds=5)
            wait(lambda: os.path.exists(os.path.join("..", "home", "certs", "watch-certs.test.crt")), timeout_seconds=5)
            wait(lambda: os.path.exists(os.path.join("..", "home", "traefik", "config", "watch-certs.test.ssl.toml")),
                 timeout_seconds=5)

            ssl_configuration = Path(
                os.path.join("..", "home", "traefik", "config", "watch-certs.test.ssl.toml")).read_text()

            assert ssl_configuration == '''# This configuration file has been automatically generated by ddb
[[tls.certificates]]
  certFile = "/certs/watch-certs.test.crt"
  keyFile = "/certs/watch-certs.test.key"
'''

            ddb_yml_content = ddb_yml_content.replace(
                'ddb.Image("httpd") + ddb.VirtualHost("80", "watch-certs.test", "httpd")',
                'ddb.Image("httpd")')

            with open("docker-compose.yml.jsonnet", "w") as f:
                f.write(ddb_yml_content)

            if not watch:
                main_runner()

            wait(lambda: not os.path.exists(os.path.join(".certs", "watch-certs.test.key")), timeout_seconds=5)
            wait(lambda: not os.path.exists(os.path.join(".certs", "watch-certs.test.crt")), timeout_seconds=5)
            wait(lambda: not os.path.exists(os.path.join("..", "home", "certs", "watch-certs.test.key")),
                 timeout_seconds=5)
            wait(lambda: not os.path.exists(os.path.join("..", "home", "certs", "watch-certs.test.crt")),
                 timeout_seconds=5)
            wait(lambda: not os.path.exists(
                os.path.join("..", "home", "traefik", "config", "watch-certs.test.ssl.toml")),
                 timeout_seconds=5)
        finally:
            if watch:
                stop_watch()

        if thread:
            thread.join()



import os
from pathlib import Path

from ddb.__main__ import load_registered_features
from ddb.config import config
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.docker import DockerFeature
from ddb.feature.jsonnet import JsonnetFeature
from ddb.feature.traefik import TraefikFeature
from ddb.feature.traefik.actions import TraefikInstalllCertsAction, TraefikUninstalllCertsAction, \
    TraefikExtraServicesAction


class TestTraefikFeature:
    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(TraefikFeature())
        load_registered_features()

        assert config.data.get('traefik.disabled') is True

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(TraefikFeature())
        load_registered_features()

        assert config.data.get('traefik.disabled') is True

    def test_install_certs(self, project_loader):
        project_loader("install-certs")

        features.register(CoreFeature())
        features.register(TraefikFeature())
        load_registered_features()

        install_action = TraefikInstalllCertsAction()
        install_action.execute(domain="dummy.tld",
                               private_key=os.path.join(".certs", "some-dummy.tld.key"),
                               certificate=os.path.join(".certs", "some-dummy.tld.crt"))

        assert os.path.exists(os.path.join(config.paths.home, "certs", "dummy.tld.key"))
        assert os.path.exists(os.path.join(config.paths.home, "certs", "dummy.tld.crt"))

        with open(os.path.join(config.paths.home, "certs", "dummy.tld.crt")) as crt_file:
            assert 'crt' == crt_file.read()

        with open(os.path.join(config.paths.home, "certs", "dummy.tld.key")) as key_file:
            assert 'key' == key_file.read()

        assert os.path.exists(os.path.join(config.data['traefik']['config_directory'], "dummy.tld.ssl.toml"))

        uninstall_action = TraefikUninstalllCertsAction()
        uninstall_action.execute(domain="dummy.tld")

        assert not os.path.exists(os.path.join(config.data['traefik']['config_directory'], "dummy.tld.ssl.toml"))

    def test_extra_services(self, project_loader):
        project_loader("extra-services")

        features.register(CoreFeature())
        features.register(TraefikFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()

        install_action = TraefikExtraServicesAction()
        install_action.initialize()
        install_action.execute()

        files = {
            "api": {
                "generated": "sub.project.test.extra-service.api.toml",
                "expected": "sub.project.test.extra-service.api.expected.toml"
            },
            "rule": {
                "generated": "rule.project.test.extra-service.web.toml",
                "expected": "rule.project.test.extra-service.web.expected.toml",
                "replaces": {
                    "{{ip}}": config.data.get('docker.debug.host')
                }
            },
            "secured": {
                "generated": "secured.project.test.extra-service.redirect.toml",
                "expected": "secured.project.test.extra-service.redirect.expected.toml"
            },
            "secured_path_prefix": {
                "generated": "secured.project.test.extra-service.path_prefix.toml",
                "expected": "secured.project.test.extra-service.path_prefix.expected.toml"
            },
            "secured_path_prefix_with_redirect": {
                "generated": "secured.project.test.extra-service.path_prefix_with_redirect.toml",
                "expected": "secured.project.test.extra-service.path_prefix_with_redirect.expected.toml"
            }
        }

        for file in files:
            generated = os.path.join(config.data['traefik']['config_directory'], files.get(file).get('generated'))
            assert os.path.exists(generated)
            expected = Path(
                os.path.join(config.data['traefik']['config_directory'], files.get(file).get('expected'))
            ).read_text()
            if files.get(file).get('replaces'):
                for replace in files.get(file).get('replaces'):
                    expected = expected.replace(replace, files.get(file).get('replaces').get(replace))
            assert expected == Path(generated).read_text()

    def test_extra_services_template(self, project_loader):
        project_loader("extra-services-template")

        features.register(CoreFeature())
        features.register(TraefikFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()

        install_action = TraefikExtraServicesAction()
        install_action.initialize()
        install_action.execute()

        api_toml = os.path.join(config.data['traefik']['config_directory'],
                                "sub.project.test.extra-service.api.toml")
        assert os.path.exists(api_toml)

        api_toml_expected = Path(os.path.join(config.data['traefik']['config_directory'],
                                              "sub.project.test.extra-service.api.expected.toml"))

        assert api_toml_expected.read_text() == Path(api_toml).read_text()

        web_toml = os.path.join(config.data['traefik']['config_directory'], "rule.project.test.extra-service.web.toml")
        assert os.path.exists(web_toml)

        web_toml_expected = Path(os.path.join(config.data['traefik']['config_directory'],
                                              "rule.project.test.extra-service.web.expected.toml"))

        assert web_toml_expected.read_text().replace('{{ip}}', config.data.get('docker.debug.host')) == Path(
            web_toml).read_text()

        secured_toml = os.path.join(config.data['traefik']['config_directory'],
                                    "secured.project.test.extra-service.redirect.toml")
        assert os.path.exists(secured_toml)

        secured_toml_expected = Path(os.path.join(config.data['traefik']['config_directory'],
                                                  "secured.project.test.extra-service.redirect.expected.toml"))

        assert secured_toml_expected.read_text() == Path(secured_toml).read_text()

    def test_extra_services_redirect(self, project_loader):
        project_loader("extra-services-redirect")

        features.register(CoreFeature())
        features.register(TraefikFeature())
        features.register(DockerFeature())
        features.register(JsonnetFeature())
        load_registered_features()

        install_action = TraefikExtraServicesAction()
        install_action.initialize()
        install_action.execute()

        api_toml = os.path.join(config.data['traefik']['config_directory'],
                                "sub.project.test.extra-service.api.toml")
        assert os.path.exists(api_toml)

        api_toml_expected = Path(os.path.join(config.data['traefik']['config_directory'],
                                              "sub.project.test.extra-service.api.expected.toml"))

        assert api_toml_expected.read_text() == Path(api_toml).read_text()

        web_toml = os.path.join(config.data['traefik']['config_directory'], "rule.project.test.extra-service.web.toml")
        assert os.path.exists(web_toml)

        web_toml_expected = Path(os.path.join(config.data['traefik']['config_directory'],
                                              "rule.project.test.extra-service.web.expected.toml"))

        assert web_toml_expected.read_text().replace('{{ip}}', config.data.get('docker.debug.host')) == Path(
            web_toml).read_text()

        secured_toml = os.path.join(config.data['traefik']['config_directory'],
                                    "secured.project.test.extra-service.no-redirect.toml")
        assert os.path.exists(secured_toml)

        secured_toml_expected = Path(os.path.join(config.data['traefik']['config_directory'],
                                                  "secured.project.test.extra-service.no-redirect.expected.toml"))

        assert secured_toml_expected.read_text() == Path(secured_toml).read_text()



# -*- coding: utf-8 -*-
from ddb.feature.schema import FeatureSchema
from marshmallow import fields, Schema


class ExtraServiceSchema(Schema):
    """
    Extra Service
    """
    https = fields.Bool(required=False, allow_none=True, default=None)
    url = fields.String(required=True, allow_none=False, default=None)
    domain = fields.String(required=True, allow_none=True, default=None)
    rule = fields.String(required=False, allow_none=True, default=None)
    redirect_to_https = fields.Boolean(required=False, allow_none=True, default=None)
    path_prefix = fields.String(required=False, allow_none=True, default=None)
    redirect_to_path_prefix = fields.Boolean(required=False, allow_none=True, default=None)


class TraefikSchema(FeatureSchema):
    """
    Traefik feature schema.
    """
    certs_directory = fields.String(required=False, allow_none=True, default=None)
    config_directory = fields.String(required=False, allow_none=True, default=None)
    mapped_certs_directory = fields.String(required=True, default="/certs")
    ssl_config_template = fields.String(required=True, default="""
# This configuration file has been automatically generated by ddb
[[tls.certificates]]
  certFile = "{{_local.certFile}}"
  keyFile = "{{_local.keyFile}}"

""".lstrip())
    extra_services = fields.Dict(fields.String(), fields.Nested(ExtraServiceSchema()), default={})
    extra_services_config_template = fields.String(required=True, default="""
# This configuration file has been automatically generated by ddb
[http.routers]
{%- if _local.https is none or _local.https is sameas false %}
  [http.routers.extra-service-{{_local.id}}]
    rule = "{{_local.rule}}"{% if not _local.redirect_to_https and _local.path_prefix %} && "PathPrefix(`{{_local.path_prefix}}{regex:$$|/.*}`)"{% endif %}
    entrypoints = ["http"]
    service = "extra-service-{{_local.id}}"
{%- if _local.redirect_to_https %}
    middlewares = ["extra-service-{{_local.id}}-redirect-to-https"]
{%- elif _local.path_prefix %}
    middlewares = ["extra-service-{{_local.id}}-stripprefix"]
{%- endif %}
{%- endif %}
{%- if _local.https is none or _local.https is sameas true %}
  [http.routers.extra-service-{{_local.id}}-tls]
    rule = "{{_local.rule}}{% if _local.path_prefix %} && PathPrefix(`{{_local.path_prefix}}{regex:$$|/.*}`){% endif %}"
    entrypoints = ["https"]
    tls = true
    service = "extra-service-{{_local.id}}"
{%- if _local.path_prefix %}
    middlewares = ["extra-service-{{_local.id}}-stripprefix"]
{%- endif %}
{%- if _local.certresolver is defined %}
    [http.routers.extra-service-{{_local.service}}-tls.tls]
      certResolver = "{{_local.certresolver}}"
{%- endif %}
{%- endif %}
{%- if _local.redirect_to_path_prefix %}
{%- if _local.https is none and not _local.redirect_to_https or _local.https is sameas false %}
  [http.routers.extra-service-{{_local.id}}-redirect-to-path-prefix]
    rule = "{{_local.rule}}"
    entrypoints = ["http"]
    service = "extra-service-{{_local.id}}"
    middlewares = ["extra-service-{{_local.id}}-redirect-to-path-prefix"]
{%- endif %}
{%- if _local.https is none or _local.https is sameas true %}
  [http.routers.extra-service-{{_local.id}}-redirect-to-path-prefix-tls]
    rule = "{{_local.rule}}"
    entrypoints = ["https"]
    tls = true
    service = "extra-service-{{_local.id}}"
    middlewares = ["extra-service-{{_local.id}}-redirect-to-path-prefix"]
{%- if _local.certresolver is defined %}
    [http.routers.extra-service-{{_local.service}}-tls.tls]
      certResolver = "{{_local.certresolver}}"
{%- endif %}
{%- endif %}
{%- endif %}

{%- if _local.redirect_to_https or _local.path_prefix  %}

[http.middlewares]
{%- if _local.redirect_to_https %}
  [http.middlewares.extra-service-{{_local.id}}-redirect-to-https.redirectScheme]
    scheme = "https"
{%- endif %}
{%- if _local.path_prefix %}
  [http.middlewares.extra-service-{{_local.id}}-stripprefix.stripPrefix]
    prefixes = "{{_local.path_prefix}}"
{%- endif %}
{%- if _local.redirect_to_path_prefix %}
  [http.middlewares.extra-service-{{_local.id}}-redirect-to-path-prefix.redirectregex]
    regex = "^.*$"
    replacement = "{{_local.path_prefix}}"
{%- endif %}
{%- endif %}

[http.services]
  [http.services.extra-service-{{_local.id}}]
    [http.services.extra-service-{{_local.id}}.loadBalancer]
      [[http.services.extra-service-{{_local.id}}.loadBalancer.servers]]
        url = "{{_local.url}}"

""".lstrip())



# -*- coding: utf-8 -*-
from typing import ClassVar, Iterable

from dotty_dict import Dotty

from ddb.action import Action
from ddb.feature import Feature
from .actions import CookiecutterAction
from .schema import CookiecutterFeatureSchema


class CookiecutterFeature(Feature):
    """
    Generate code from a cookiecutter template available on github (https://cookiecutter.readthedocs.io).
    """

    @property
    def name(self) -> str:
        return "cookiecutter"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[CookiecutterFeatureSchema]:
        return CookiecutterFeatureSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            CookiecutterAction(),
        )

    def _configure_defaults(self, feature_config: Dotty):
        if 'templates' in feature_config:
            feature_config['templates'] = [self._sanitize_template(t) for t in feature_config['templates']]

    @staticmethod
    def _sanitize_template(template):
        if template.get('output_dir'):
            template['extra_context'] = {'directory': '.'}
        return template



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema


class FixuidSchema(FeatureSchema):
    """
    Certs feature schema.
    """
    url = fields.String(required=True, default="https://github.com/boxboat/fixuid/releases/download/"
                                               "v0.5.1/fixuid-0.5.1-linux-amd64.tar.gz")



# -*- coding: utf-8 -*-
import os
import shutil
import zipfile
from pathlib import Path
from typing import Callable, Optional

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.tmpdir import TempPathFactory
from ddb.__main__ import reset, configure_logging
from ddb.config import Config
from pytest_mock import MockerFixture
from verboselogs import SPAM

from .utilstest import init_config_paths

pytest_plugins = ["docker_compose"]


@pytest.fixture(scope="module")
def global_data_dir(request: FixtureRequest) -> str:
    filename = request.module.__file__
    dirname = os.path.dirname(filename)
    return os.path.join(os.path.join(dirname, "data"))


@pytest.fixture(scope="module")
def data_dir(global_data_dir: str, request: FixtureRequest) -> str:
    filename = request.module.__file__
    dirname = os.path.dirname(global_data_dir)
    data_dirname, _ = os.path.splitext(os.path.basename(filename))
    return os.path.join(dirname, data_dirname + ".data")


@pytest.fixture()
def project_loader(data_dir: str, tmp_path_factory: TempPathFactory, request: FixtureRequest) -> Callable[
    [Optional[str]], Config]:
    def load(name: str = None, before_load_config=None, config_provider=init_config_paths):
        root_dir = os.path.join(data_dir, name) if name else data_dir

        tmp_path = tmp_path_factory.mktemp(request.function.__name__)  # type: Path
        tmp_path.rmdir()
        shutil.copytree(root_dir, str(tmp_path))

        os.chdir(str(tmp_path))
        if before_load_config:
            before_load_config()
        conf = config_provider(str(tmp_path))
        os.chdir(str(conf.paths.project_home))

        if os.path.exists("repo.zip"):
            with zipfile.ZipFile("repo.zip", 'r') as zip_ref:
                zip_ref.extractall(".")

        return conf

    return load


@pytest.fixture(autouse=True)
def configure(mocker: MockerFixture):
    original_environ = dict(os.environ)
    cwd = os.getcwd()

    Config.defaults = {'defaults': {'fail_fast': True}}

    def overrides(config):
        if not 'core' in config:
            config['core'] = {}
        config.get('core')['check_updates'] = False

    Config.overrides = overrides

    try:
        if os.name == 'nt' and 'COMSPEC' not in os.environ:
            os.environ['COMSPEC'] = r'C:\Windows\System32\cmd.exe'
        if os.name != 'nt' and 'SHELL' not in os.environ:
            os.environ['SHELL'] = '/bin/bash'
        if os.name == 'nt':
            os.environ['DDB_OVERRIDE_DOCKER_USER_UID'] = '1000'
            os.environ['DDB_OVERRIDE_DOCKER_USER_GID'] = '1000'
            os.environ['DDB_OVERRIDE_DOCKER_IP'] = '127.0.0.1'

        configure_logging(SPAM)

        mocker.patch('ddb.feature.smartcd.actions.is_smartcd_installed', lambda: False)
        mocker.patch('ddb.feature.version.is_git_repository', lambda: False)

        yield
    finally:
        os.chdir(cwd)
        os.environ.clear()
        os.environ.update(original_environ)
        reset()


def pytest_configure(config):
    """
    Makes verbose pytest diff without -v flag.

    https://stackoverflow.com/questions/30938780/how-can-i-show-verbose-py-test-diffs-without-verbose-test-progress
    """
    terminal = config.pluginmanager.getplugin('terminal')

    class QuietReporter(terminal.TerminalReporter):
        @property
        def verbosity(self):
            return 0

        @property
        def showlongtestinfo(self):
            return False

        @property
        def showfspath(self):
            return False

    terminal.TerminalReporter = QuietReporter



from _pytest.logging import LogCaptureFixture

from ddb.__main__ import main
from ddb.config import Config


class TestErrorHandling:
    def test_invalid_jsonnet(self, project_loader, caplog: LogCaptureFixture):
        Config.defaults = None

        project_loader("invalid-jsonnet")

        exceptions = main(["configure"])
        assert len(exceptions) == 1

        assert len(caplog.records) >= 1
        record = caplog.records[0]

        assert record.message == \
               'An unexpected error has occured ' + \
               '[phase:configure => FileWalkAction.execute(), ' + \
               'file:found => JsonnetAction.execute(target=invalid, template=invalid.jsonnet)]: ' + \
               'STATIC ERROR: invalid.jsonnet:1:1-9: Unknown variable: trololol'



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema


class VersionSchema(FeatureSchema):
    """
    Git schema.
    """
    branch = fields.String(allow_none=True)  # default is set in feature _configure_defaults
    tag = fields.String(allow_none=True)  # default is set in feature _configure_defaults
    version = fields.String(allow_none=True)  # default is set in feature _configure_defaults
    hash = fields.String(allow_none=True)  # default is set in feature _configure_defaults
    short_hash = fields.String(allow_none=True)  # default is set in feature _configure_defaults



from ddb.action import Action
from ddb.event import events


class CustomAction(Action):
    @property
    def event_bindings(self):
        return events.phase.configure

    @property
    def name(self):
        return "custom_action3"

    def execute(self):
        with open("test3", "w") as f:
            f.write("Custom3Action")



import os

import yaml
from _pytest.capture import CaptureFixture
from dotty_dict import Dotty

from ddb.__main__ import main, reset
from ddb.config import config


class TestConfig:
    def test_config_override_shell(self, project_loader):
        project_loader("override-shell")

        os.environ['DDB_OVERRIDE_SHELL_SHELL'] = 'fish'

        main(["config"], reset_disabled=True)

        assert config.data.get('shell.shell') == 'fish'

        reset()

    def test_config_override_ci(self, project_loader):
        project_loader("override-ci")

        os.chdir("sub")

        main(["config"], reset_disabled=True)

        assert config.data.get('core.env.current') == 'ci'
        assert config.data.get('docker.cache_from_image') is True
        assert config.data.get('docker.registry.name') == "gfiorleans.azurecr.io"
        assert config.data.get('docker.registry.repository') == "alm-atout"

        reset()

    def test_config_output_extra_filenames(self, project_loader, capsys: CaptureFixture):
        project_loader("extra-filenames")

        main(["config"])

        configuration = Dotty(yaml.safe_load(capsys.readouterr().out))

        assert configuration['app.value'] == 'local'
        assert configuration['some'] is True
        assert configuration['another'] is True
        assert configuration['core.configuration.extra'] == ['some.custom.yml', 'another.config.file']

        reset()

    def test_config_output_extra_filenames_files_option(self, project_loader, capsys: CaptureFixture):
        project_loader("extra-filenames")

        main(["config", "--files"])

        reset()

        output = capsys.readouterr().out

        parts = [part.lstrip() for part in output.split('---') if part.strip()]
        assert len(parts) == 4

        configurations = {}

        for part in parts:
            filename, config = part.split('\n', 1)
            assert filename.startswith('# ')
            filename = filename[2:]
            filename = os.path.relpath(filename, os.getcwd())
            configurations[filename] = Dotty(yaml.safe_load(config))

        assert ('ddb.yml', 'some.custom.yml', 'another.config.file.yaml', 'ddb.local.yml') == \
               tuple(configurations.keys())

        assert configurations['ddb.yml']['app.value'] == 'default'
        assert configurations['some.custom.yml']['app.value'] == 'custom'
        assert configurations['another.config.file.yaml']['app.value'] == 'config'
        assert configurations['ddb.local.yml']['app.value'] == 'local'

    def test_config_output_extra_filenames_some_files_option(self, project_loader, capsys: CaptureFixture):
        project_loader("extra-filenames")

        main(["config", "some", "--files"])

        reset()

        output = capsys.readouterr().out

        parts = [part.lstrip() for part in output.split('---') if part.strip()]
        assert len(parts) == 1

        configurations = {}

        for part in parts:
            filename, config = part.split('\n', 1)
            assert filename.startswith('# ')
            filename = filename[2:]
            filename = os.path.relpath(filename, os.getcwd())
            configurations[filename] = Dotty(yaml.safe_load(config))

        assert ('some.custom.yml',) == \
               tuple(configurations.keys())

        assert configurations['some.custom.yml']['some'] is True
        assert 'app.value' not in configurations['some.custom.yml']

    def test_config_local_falsy(self, project_loader):
        project_loader("local-falsy")

        main(["configure"], reset_disabled=True)

        assert config.data.get('app.disabled_services') == []

        reset()

    def test_config_custom_strategy(self, project_loader):
        project_loader("local-custom-strategy")

        main(["configure"], reset_disabled=True)

        assert config.data.get('app.disabled_services') == ['python', 'gunicorn', 'another']
        assert config.data.get('app.another_strategy') == ['another', 'python', 'gunicorn']

        reset()

    def test_config_deep_custom_strategy(self, project_loader):
        project_loader("deep-local-custom-strategy")

        main(["configure"], reset_disabled=True)

        assert config.data.get('app.deep.disabled_services') == ['python', 'gunicorn', 'another']
        assert config.data.get('app.deep.another_strategy') == ['another', 'python', 'gunicorn']

        reset()

    def test_config_env_ddb(self, project_loader):
        project_loader("env-ddb")

        main(["configure"], reset_disabled=True)

        assert not config.data.get('app.test')

        reset()

    def test_config_env_ddb2(self, project_loader):
        project_loader("env-ddb")

        os.rename('ddb.dev.tmp.yml', 'ddb.dev.yml')

        main(["configure"], reset_disabled=True)

        assert config.data.get('app.test')
        assert not os.path.islink("ddb.yml")

        reset()

    def test_config_extra_filenames(self, project_loader):
        project_loader("extra-filenames")

        main(["configure"], reset_disabled=True)

        assert config.data.get('another') is True
        assert config.data.get('some') is True
        assert config.data.get('app.value') == 'local'

        reset()

    def test_config_merge_default(self, project_loader):
        project_loader("merge-default")

        main(["configure"], reset_disabled=True)

        assert config.data.get('core.env.current') == 'dev-services'
        assert config.data.get('core.env.available') == ['prod', 'stage', 'ci', 'dev', 'dev-services']

        reset()

    def test_config_merge_insert_strategy(self, project_loader):
        project_loader("merge-insert-strategy")

        main(["configure"], reset_disabled=True)

        assert config.data.get('core.env.current') == 'dev'
        assert config.data.get('core.env.available') == ['prod', 'stage', 'ci', 'dev-services', 'dev-services2', 'dev']

        reset()

    def test_config_merge_insert_strategy2(self, project_loader):
        project_loader("merge-insert-strategy2")

        main(["configure"], reset_disabled=True)

        assert config.data.get('core.env.current') == 'dev'
        assert config.data.get('core.env.available') == ['prod', 'prod-services', 'prod-services2', 'stage', 'ci',
                                                         'dev']

        reset()

    def test_config_merge_insert_strategy3(self, project_loader):
        project_loader("merge-insert-strategy3")

        main(["configure"], reset_disabled=True)

        assert config.data.get('core.env.current') == 'dev'
        assert config.data.get('core.env.available') == ['prod', 'prod', 'stage', 'stage', 'ci',
                                                         'dev']

        reset()

    def test_config_more_properties_jsonnet_docker(self, project_loader, capsys: CaptureFixture):
        project_loader("more-properties")

        main(["config", "jsonnet.docker"])

        configuration = Dotty(yaml.safe_load(capsys.readouterr().out))

        assert configuration['jsonnet.docker.compose.project_name'] == 'yo-custom'
        assert configuration['jsonnet.docker.registry.name'] == 'gfiorleans.azurecr.io'
        assert configuration['jsonnet.docker.registry.repository'] == 'yo-custom'
        assert configuration['jsonnet.docker.virtualhost.redirect_to_https'] is True

        assert 'docker' not in configuration
        assert 'core' not in configuration

        reset()

    def test_config_more_properties_jsonnet_docker_compose(self, project_loader, capsys: CaptureFixture):
        project_loader("more-properties")

        main(["config", "jsonnet.docker.compose"])

        configuration = Dotty(yaml.safe_load(capsys.readouterr().out))

        assert configuration['jsonnet.docker.compose.project_name'] == 'yo-custom'
        assert 'jsonnet.docker.registry.name' not in configuration
        assert 'jsonnet.docker.registry.repository' not in configuration
        assert 'jsonnet.docker.virtualhost.redirect_to_https' not in configuration
        assert 'docker' not in configuration
        assert 'core' not in configuration

        reset()

    def test_config_more_properties_docker_variables_full(self, project_loader, capsys: CaptureFixture):
        project_loader("more-properties")

        main(["config", "docker", "--variables", "--full"], reset_disabled=True)

        out = capsys.readouterr().out

        docker_ip = config.data.get('docker.ip')
        docker_interface = config.data.get('docker.interface')
        docker_user_gid = config.data.get('docker.user.gid')
        docker_user_uid = config.data.get('docker.user.uid')

        assert out == f"""docker.disabled: False
docker.interface: {docker_interface}
docker.ip: {docker_ip}
docker.user.gid: {docker_user_gid}
docker.user.group: None
docker.user.name: None
docker.user.uid: {docker_user_uid}
"""

        reset()

    def test_config_more_properties_docker_user_variables(self, project_loader, capsys: CaptureFixture):
        project_loader("more-properties")

        main(["config", "docker.user", "--variables"], reset_disabled=True)

        out = capsys.readouterr().out

        docker_user_gid = config.data.get('docker.user.gid')
        docker_user_uid = config.data.get('docker.user.uid')

        assert out == f"""docker.user.gid: {docker_user_gid}
docker.user.group: None
docker.user.name: None
docker.user.uid: {docker_user_uid}
"""

        reset()

    def test_config_more_properties_docker_ip_value(self, project_loader, capsys: CaptureFixture):
        project_loader("more-properties")

        main(["config", "docker.ip", "--value"], reset_disabled=True)

        out = capsys.readouterr().out

        docker_ip = config.data.get('docker.ip')

        assert out == f"{docker_ip}\n"

        reset()

    def test_config_more_properties_core_env_available_value(self, project_loader, capsys: CaptureFixture):
        project_loader("more-properties")

        main(["config", "core.env.available", "--value"], reset_disabled=True)

        out = capsys.readouterr().out

        available = config.data.get('core.env.available')

        assert out == f"{available}\n"

        reset()



# -*- coding: utf-8 -*-
from typing import Iterable, ClassVar

from ddb.action import Action
from ddb.feature import Feature
from .actions import SmartcdAction, WindowsProjectActivate


class SmartcdFeature(Feature):
    """
    Smartcd support (https://github.com/cxreg/smartcd).
    """

    @property
    def name(self) -> str:
        return "smartcd"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def actions(self) -> Iterable[Action]:
        return (
            SmartcdAction(),
            WindowsProjectActivate(),
        )



# -*- coding: utf-8 -*-
from typing import ClassVar, Iterable

from dotty_dict import Dotty

from .actions import SymlinksAction
from .schema import SymlinksSchema
from ..feature import Feature, FeatureConfigurationError
from ...action import Action
from ...config import config
from ...utils.file import TemplateFinder, FileWalker


class SymlinksFeature(Feature):
    """
    Creates symlinks from file matching current environment.
    """

    @property
    def name(self) -> str:
        return "symlinks"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core", "file"]

    @property
    def schema(self) -> ClassVar[SymlinksSchema]:
        return SymlinksSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            SymlinksAction(),
        )

    def _configure_defaults(self, feature_config: Dotty):
        suffixes = feature_config.get('suffixes')
        if not suffixes:
            try:
                env_available = config.data["core.env.available"]
            except KeyError as err:
                raise FeatureConfigurationError(self,
                                                "core.env.available or symlinks.suffixes should be defined.") \
                    from err

            available_suffixes = ["." + env for env in env_available]

            env_current = config.data.get("core.env.current")
            if not env_current:
                current_suffix = available_suffixes[-1]
            else:
                current_suffix = "." + env_current

            suffixes = available_suffixes[available_suffixes.index(current_suffix):]
            feature_config["suffixes"] = suffixes

        includes = feature_config.get('includes')
        if includes is None:
            includes = FileWalker.build_default_includes_from_suffixes(suffixes)
            feature_config["includes"] = includes



# -*- coding: utf-8 -*-
from typing import Optional, Iterable

from ...action import Action
from ...action.action import EventBinding
from ...binary import binaries, Binary
from ...config import config
from ...event import events


class RunAction(Action):
    """
    Run a binary
    """

    @property
    def event_bindings(self):
        return (
            events.phase.run,
            EventBinding(events.run.run, self.run)
        )

    @property
    def name(self) -> str:
        return "run:run"

    @staticmethod
    def execute():
        """
        Execute the action defined by argument "name".
        """
        if hasattr(config.args, "name"):
            name = config.args.name
            events.run.run(name=name)

    @staticmethod
    def run(name: str, *args):
        """
        Execute the action
        """
        if not args:
            args = config.unknown_args

        binary_set = binaries.get(name)
        binary = None
        if binary_set:
            binary = RunAction.select_binary(binary_set, *args)

        if binary:
            binary.before_run()
            events.run.command(binary.command(*args), False)
        else:
            events.run.command([name, *args], True)

    @staticmethod
    def select_binary(binary_set: Iterable[Binary], *args) -> Optional[Binary]:
        """
        Select effective binary between many binaries.
        """
        ordered_binary = sorted(binary_set)
        for binary in ordered_binary:
            if binary.should_run(*args):
                return binary
        return None



import os

import pytest
import yaml
from dotty_dict import Dotty

from ddb.__main__ import main
from ddb.config import config
from tests.utilstest import get_user_uid_gid, get_group_gid


class TestJsonnet:
    def test_jsonnet_ext_var(self, project_loader):
        project_loader("jsonnet-extvar")

        main(["configure"])

        assert os.path.exists('config.yml')
        with open('config.yml', 'r') as f:
            variables = f.read()

        with open('config.expected.yml', 'r') as f:
            variables_expected = f.read()

        assert variables == variables_expected


class TestDockerJsonnet:
    @pytest.mark.skipif("os.name == 'nt'")
    def test_named_user_group(self, project_loader):
        project_loader("named-user-group")

        main(["configure"], reset_disabled=True)

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            docker_compose = yaml.load(f, yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))
            docker_compose_expected = yaml.load(expected_data, yaml.SafeLoader)

        uid, _ = get_user_uid_gid('root')
        gid = get_group_gid('nobody')

        Dotty(docker_compose_expected)['services.maven.user'] = f"{uid}:{gid}"

        assert docker_compose == docker_compose_expected

    def test_named_invalid_user_group(self, project_loader):
        project_loader("named-invalid-user-group")

        with pytest.raises(RuntimeError):
            main(["configure"])

    def test_binary(self, project_loader):
        project_loader("jsonnet-binary")

        main(["configure"], reset_disabled=True)

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            docker_compose = yaml.load(f, yaml.SafeLoader)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))

            docker_compose_expected = yaml.load(expected_data, yaml.SafeLoader)

        assert docker_compose == docker_compose_expected

    def test_resolve_ports_conflicts(self, project_loader):
        project_loader("jsonnet-resolve-ports-conflicts")

        main(["configure"], reset_disabled=True)

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            docker_compose = yaml.safe_load(f)

        with open('docker-compose.expected.yml', 'r') as f:
            expected_data = f.read()
            expected_data = expected_data.replace("%network_name%",
                                                  str(config.data.get('jsonnet.docker.compose.network_name')))

            docker_compose_expected = yaml.load(expected_data, yaml.SafeLoader)

        assert docker_compose == docker_compose_expected



# -*- coding: utf-8 -*-
import fnmatch
import os
import re
import shutil
from pathlib import Path
from tempfile import gettempdir, NamedTemporaryFile
from typing import List, Union, Optional, Tuple

import chmod_monkey
import requests
from braceexpand import braceexpand

from ddb.config import config
from ddb.context import context


def has_same_content(filename1: str, filename2: str, read_mode='rb') -> bool:
    """
    Check if the content of two files are same
    """
    with open(filename1, mode=read_mode) as file1:  # pylint:disable=unspecified-encoding
        with open(filename2, mode=read_mode) as file2:  # pylint:disable=unspecified-encoding
            return file1.read() == file2.read()


def write_if_different(file, data, read_mode='r', write_mode='w', log_source=None, readonly=True, **kwargs) -> bool:
    """
    Write the file if existing data is different than given data.
    """
    try:
        read_encoding = "utf-8" if 'b' not in read_mode else None  # TODO: make default encoding configurable
        write_encoding = "utf-8" if 'b' not in write_mode else None

        if os.path.exists(file):
            try:
                with open(file, mode=read_mode, encoding=read_encoding, **kwargs) as read_file:
                    existing_data = read_file.read()
            except OSError:
                shutil.rmtree(file)
                existing_data = None
        else:
            existing_data = None

        if existing_data == data:
            context.log.notice("%s -> %s", log_source if log_source else "", file)
            return False

        if not os.access(file, os.W_OK) and os.path.isfile(file):
            chmod(file, '+w', logging=False)
        with open(file, mode=write_mode, encoding=write_encoding, **kwargs) as write_file:
            write_file.write(data)

        context.log.success("%s -> %s", log_source if log_source else "", file)

        return True
    finally:
        if readonly and os.path.isfile(file):
            chmod(file, '+r', logging=False)


def copy_if_different(source, target, read_mode='r', write_mode='w', log=False, readonly=True, **kwargs) -> bool:
    """
    Copy source to target if existing source data is different than target data.
    """
    try:
        read_encoding = "utf-8" if 'b' not in read_mode else None  # TODO: make default encoding configurable
        write_encoding = "utf-8" if 'b' not in write_mode else None

        if os.path.exists(source):
            with open(source, mode=read_mode, encoding=read_encoding, **kwargs) as read_file:
                source_data = read_file.read()
        else:
            source_data = None

        if os.path.exists(target):
            with open(target, mode=read_mode, encoding=read_encoding, **kwargs) as read_file:
                target_data = read_file.read()
        else:
            target_data = None

        if source_data == target_data:
            if log:
                context.log.notice("%s -> %s", source, target)
            return False

        if not os.access(target, os.W_OK) and os.path.exists(target):
            chmod(target, '+w')
        with open(target, mode=write_mode, encoding=write_encoding, **kwargs) as write_file:
            write_file.write(source_data)

        if log:
            context.log.success("%s -> %s", source, target)

        return True
    finally:
        if readonly and os.path.isfile(target):
            chmod(target, '+r', logging=False)


def force_remove(file: str, silent=False):
    """
    Remove a file, trying to unset readonly flag if a PermissionError occurs.
    """
    try:
        os.remove(file)
    except PermissionError:
        # On windows, removing a readonly file raise this error.
        chmod(file, '+w', logging=False)
        os.remove(file)
    except FileNotFoundError:
        if not silent:
            context.log.warning("%s can't be removed because it's already absent", file)


def chmod(file: str, mode: str, logging=True):
    """
    Apply given mode to file
    """
    new_mode, old_mode = chmod_monkey.to_mode(file, mode, return_old_mode=True)
    if old_mode != new_mode:
        if logging:
            context.log.success("chmod %s %s", mode, file)
        os.chmod(file, new_mode)
    else:
        if logging:
            context.log.notice("chmod %s %s", mode, file)


class FileUtils:
    """
    Some file management functions
    """

    @staticmethod
    def get_file_content(url: str) -> str:
        """
        Get the content of the file
        :param url: the path to the file (https? or file)
        :return:
        """
        if url.startswith('file://'):
            return FileUtils._get_local_file_content(url)
        return requests.get(url).text

    @staticmethod
    def _get_local_file_content(file_path: str) -> str:
        """
        Get the content of the file
        :param file_path: the path to the file (absolute or relative)
        :return:
        """
        file_path = file_path.replace('file://', '')
        if not os.path.isabs(file_path):
            file_path = os.path.join(config.path.project_home, file_path)
        with open(file_path, 'rb') as file:
            return file.read().decode("utf-8")


class FileWalker:
    """
    Walk files inside project directory.
    TODO: Add a excludes_from_gitignore option to automatically exclude files based on gitignore behavior.
    """

    # pylint:disable=too-many-arguments
    def __init__(self,
                 includes: Optional[List[str]],
                 excludes: Optional[List[str]],
                 suffixes: Optional[List[str]],
                 rootpath: Optional[Union[Path, str]] = None,
                 recursive=True,
                 skip_processed_sources=True,
                 skip_processed_targets=True):
        if includes is None:
            includes = []
        if excludes is None:
            excludes = []
        includes, excludes = self._braceexpand(includes, excludes)
        self.includes = list(map(lambda x: re.compile(fnmatch.translate(x)), includes))
        self.excludes = list(map(lambda x: re.compile(fnmatch.translate(x)), excludes))
        self.suffixes = suffixes if suffixes is not None else []
        if not rootpath:
            rootpath = os.path.relpath(config.paths.project_home)
        self.rootpath = rootpath if isinstance(rootpath, Path) else Path(rootpath)
        self.recursive = recursive
        self.skip_processed_sources = skip_processed_sources
        self.skip_processed_targets = skip_processed_targets

    @property
    def items(self):
        """
        Yields over a tuple (template, target) with found templates.
        """

        for source in self._walk(str(self.rootpath), recursive=self.recursive):
            yield from self._do_yield(source)

    def _do_yield(self, source):  # pylint:disable=no-self-use
        yield source

    def _walk(self, *args, recursive=True, **kwargs):
        _walk_generator = os.walk(*args, **kwargs)
        for root, dirs, files in os.walk(*args, **kwargs):
            for dirs_item in list(dirs):
                dirpath = os.path.relpath(os.path.join(root, dirs_item))
                if self._is_excluded(dirpath, *self.excludes):
                    dirs.remove(dirs_item)
                else:
                    yield dirpath

            for files_item in list(files):
                filepath = os.path.relpath(os.path.join(root, files_item))
                if self._is_included(filepath, *self.includes) and \
                        not self._is_excluded(filepath, *self.excludes):
                    yield filepath

            if not recursive:
                break

    @staticmethod
    def _braceexpand(includes, excludes):
        expanded_includes = []
        for include in includes:
            expanded_includes.extend(braceexpand(include))

        expanded_excludes = []
        for exclude in excludes:
            expanded_excludes.extend(braceexpand(exclude))

        return expanded_includes, expanded_excludes

    @staticmethod
    def _is_excluded(candidate: str, *excludes: List[str]) -> bool:
        excluded = False
        if not excludes:
            return False
        # TODO find a better solution instead of adding the "./" in front of the normed path
        norm_candidate = FileWalker._prefix_path_to_current_folder(Path(os.path.normpath(candidate)).as_posix())
        for exclude in excludes:
            if exclude.match(candidate) or exclude.match(norm_candidate):
                excluded = True
                break
        return excluded

    @staticmethod
    def _has_ancestor_excluded(candidate_path: Path, *excludes: List[str]) -> bool:
        while True:
            candidate_path_parent = candidate_path.parent
            if not candidate_path_parent or candidate_path_parent == candidate_path:
                break
            candidate_path = candidate_path_parent
            if FileWalker._is_excluded(str(candidate_path), *excludes):
                return True
        return False

    @staticmethod
    def _prefix_path_to_current_folder(path: str):
        if path[0:2] == './':
            return path
        return './' + path

    def is_source_filtered(self, candidate: str):
        """
        Check if a source path is filtered out by includes/excludes
        """
        return not self._is_included(candidate, *self.includes) or \
               self._is_excluded(candidate, *self.excludes) or \
               self._has_ancestor_excluded(Path(candidate), *self.excludes)

    @staticmethod
    def _is_included(candidate: str, *includes: List[str]) -> bool:
        included = False
        norm_candidate = None
        if not includes:
            return True
        for include in includes:
            if not norm_candidate:
                norm_candidate = Path(os.path.normpath(candidate)).as_posix()
            if include.match(candidate) or include.match(norm_candidate):
                included = True
                break
        return included

    @staticmethod
    def build_default_includes_from_suffixes(suffixes: List[str], extensions=(".*", "")):
        """
        Build default includes configuration from suffixes configuration.
        """
        if extensions:
            extensions_pattern = "{" + ",".join(extensions) + "}"
        else:
            extensions_pattern = ""

        if len(suffixes) > 1:
            joined_suffixes = ','.join(suffixes)
            return ["*{" + joined_suffixes + "}" + extensions_pattern]
        if len(suffixes) > 0:
            return ["*" + suffixes[0] + extensions_pattern]
        if extensions_pattern:
            return ["*" + suffixes[0] + extensions_pattern]
        return []


class TemplateFinder(FileWalker):
    """
    Find templates sources inside project directory.
    """

    def _do_yield(self, source):
        target = self.get_target(source, check=False)
        if target:
            if self.skip_processed_targets:
                if target in context.processed_targets:
                    return
            yield source, target

    def get_target(self, source, check=True):
        """
        Get the target of given source, or None if it doesn't match suffixes.
        """
        if check:
            if self.is_source_filtered(source):
                return None

        if self.skip_processed_sources:
            if source in context.processed_sources:
                return None

        target, suffix = self._get_target_and_suffix(source, self.suffixes)

        if self.skip_processed_targets:
            if target in context.processed_targets:
                previous_source = context.processed_targets[target]
                _, previous_suffix = self._get_target_and_suffix(previous_source, self.suffixes)
                if previous_suffix not in self.suffixes or \
                        self.suffixes.index(previous_suffix) <= self.suffixes.index(suffix):
                    return None

        return target

    @staticmethod
    def _get_target_and_suffix(template_candidate: str, suffixes: List[str]) -> Optional[Tuple[str, str]]:
        basename, ext = os.path.splitext(template_candidate)
        if ext in suffixes:
            return basename, ext

        if not ext and basename.startswith("."):
            ext = basename
            basename = ""

        for suffix in suffixes:
            if basename.endswith(suffix):
                return template_candidate[:len(basename) - len(suffix)] + ext, suffix

        return None, None


def get_single_temporary_file_directory(*args):
    """
    Get directory for SingleTemporaryFile *args.
    """
    return os.path.join(gettempdir(), *args)


class SingleTemporaryFile:
    """
    A NamedTemporaryFile wrapper that is created in tmp subdirectory joined with *args strings, using options from
     **kwargs. It will keep only the last temporary file is subdirectory by cleaning up the directory on each call.
    """

    def __init__(self, *args, **kwargs):
        self.tempdir = get_single_temporary_file_directory(*args)
        os.makedirs(self.tempdir, exist_ok=True)

        if "dir" in kwargs:
            kwargs.pop("dir")
        if "delete" in kwargs:
            kwargs.pop("delete")

        # pylint:disable=consider-using-with
        self.named_temporary_file = NamedTemporaryFile(dir=self.tempdir, delete=False, **kwargs)

    def __enter__(self):
        return self.named_temporary_file.__enter__()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        tmp_filepath = self.named_temporary_file.name
        try:
            return self.named_temporary_file.__exit__(exc_type, exc_value, exc_traceback)
        finally:
            for previous_temporary_file in os.listdir(self.tempdir):
                previous_temporary_filepath = os.path.join(self.tempdir, previous_temporary_file)
                if previous_temporary_filepath != tmp_filepath:
                    os.remove(previous_temporary_filepath)



import re
from pathlib import Path

import pytest
from _pytest.capture import CaptureFixture
from ddb.__main__ import main


@pytest.mark.skipif("os.name == 'nt'")
class TestBashShell:
    def test_export_special_chars(self, capsys: CaptureFixture, project_loader):
        project_loader("export-special-chars")

        main(["configure"])

        main(["activate"])

        outerr = capsys.readouterr()
        script_lines = Path(outerr.out.split()[1].strip()).read_text().splitlines()
        matches = [re.match('export (.+)=.*$', x) for x in script_lines]
        matches = [x for x in matches if x]

        for match in matches:
            if match:
                assert '/' not in match.group(1)



# -*- coding: utf-8 -*-
import os
from argparse import ArgumentParser
from typing import Iterable, ClassVar

from dotty_dict import Dotty

from .actions import ActivateAction, DeactivateAction, CreateBinaryShim, CreateAliasShim, CheckActivatedAction, \
    CommandAction
from .integrations import BashShellIntegration, CmdShellIntegration
from .schema import ShellSchema
from ..feature import Feature, FeatureConfigurationAutoConfigureError
from ..schema import FeatureSchema
from ...action import Action
from ...command import LifecycleCommand, Command
from ...config import config
from ...phase import Phase, DefaultPhase


class ShellFeature(Feature):
    """
    Shell integration.
    """

    @property
    def name(self) -> str:
        return "shell"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        return ShellSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            ActivateAction(BashShellIntegration()),
            DeactivateAction(BashShellIntegration()),
            CreateBinaryShim(BashShellIntegration()),
            CreateAliasShim(BashShellIntegration()),
            CheckActivatedAction(BashShellIntegration()),
            CommandAction(BashShellIntegration()),
            ActivateAction(CmdShellIntegration()),
            DeactivateAction(CmdShellIntegration()),
            CreateBinaryShim(CmdShellIntegration()),
            CreateAliasShim(CmdShellIntegration()),
            CheckActivatedAction(CmdShellIntegration()),
            CommandAction(CmdShellIntegration())
        )

    @property
    def phases(self) -> Iterable[Phase]:
        def activate_parser(parser: ArgumentParser):
            parser.add_argument("--force", action="store_true",
                                help="Force activation when a project is already activated.")

        def deactivate_parser(parser: ArgumentParser):
            parser.add_argument("--force", action="store_true",
                                help="Force deactivation when a project is already deactivated")

        return (
            DefaultPhase("activate", "Write a shell script to be executed to activate environment",
                         activate_parser),
            DefaultPhase("deactivate", "Write a shell script to be executed to deactivate environment",
                         deactivate_parser),
            DefaultPhase("check-activated", "Check if project is activated in current shell"),
        )

    @property
    def commands(self) -> Iterable[Command]:
        return (
            LifecycleCommand("activate",
                             "Write a shell script to be executed to activate environment",
                             "activate",
                             avoid_stdout=True,
                             ),
            LifecycleCommand("deactivate",
                             "Write a shell script to be executed to deactivate environment",
                             "deactivate",
                             avoid_stdout=True
                             ),
            LifecycleCommand("check-activated",
                             "Check if project is activated in current shell",
                             "check-activated",
                             avoid_stdout=True
                             ),
        )

    def _configure_defaults(self, feature_config: Dotty):
        if not feature_config.get('shell'):
            comspec = os.environ.get('COMSPEC')
            shell = os.environ.get('SHELL')
            if comspec and comspec.endswith('cmd.exe'):
                feature_config['shell'] = 'cmd'
            elif shell and shell.endswith('bash'):
                feature_config['shell'] = 'bash'
            elif shell and shell.endswith('zsh'):
                feature_config['shell'] = 'zsh'
            else:
                raise FeatureConfigurationAutoConfigureError(self, 'shell')

        directories = feature_config.get('path.directories')
        absolute_directories = []
        for directory in directories:
            if os.path.isabs(directory):
                absolute_directories.append(directory)
            else:
                absolute_directory = os.path.join(config.paths.project_home, directory)
                absolute_directories.append(absolute_directory)
        feature_config['path.directories'] = absolute_directories



# -*- coding: utf-8 -*-
import os
from typing import Union, Iterable, Tuple

from ddb.action.action import AbstractTemplateAction
from ddb.config import config
from ddb.context import context
from ddb.utils.file import TemplateFinder, force_remove


class SymlinksAction(AbstractTemplateAction):
    """
    Creates symbolic links based on filename suffixes.
    """

    @property
    def name(self) -> str:
        return "symlinks:create"

    def _build_template_finder(self) -> TemplateFinder:
        return TemplateFinder(config.data.get("symlinks.includes"),
                              config.data.get("symlinks.excludes"),
                              config.data.get("symlinks.suffixes"))

    def _render_template(self, template: str, target: str) -> Iterable[Tuple[Union[str, bytes, bool], str]]:
        """
        Create a symbolic link
        """
        reltemplate = os.path.relpath(template, os.path.dirname(target))

        if os.path.islink(target) and os.readlink(target) == reltemplate:
            context.log.notice("%s -> %s", template, target)
            yield False, target
        else:
            if os.path.exists(target):
                force_remove(target)

            os.symlink(reltemplate, os.path.normpath(target))
            context.log.success("%s -> %s", template, target)

            yield True, target

    def _target_is_modified(self, template: str, target: str) -> bool:
        if not os.path.islink(target):
            return True
        reltemplate = os.path.relpath(template, os.path.dirname(target))
        return os.readlink(target) != reltemplate



from collections import defaultdict

from ddb.cache import register_project_cache, caches


class RemovalCacheSupport:  # pylint:disable=too-many-instance-attributes
    """
    Handle a cache to support removal of previously generated resources when they are not linked to the current
    configuration anymore.
    """

    def __init__(self, cache_name, keys):
        self.cache_name = cache_name
        self.keys = set(keys)
        self.type = set

        self.cached = defaultdict(self.type)
        self.previous = defaultdict(self.type)
        self.current = defaultdict(self.type)

        register_project_cache(self.cache_name)

    def prepare(self):
        """
        This should be called before processing extra-services configuration.
        """
        cache = caches.get(self.cache_name)

        for key in self.keys:
            self.cached[key] = cache.get(key, self.type())
            self.previous[key] = self.type(self.cached[key])
            self.current[key] = self.type()

    def set_current_value(self, key, value):
        """
        Set current value for given key.
        """
        if key not in self.keys:
            raise ValueError("Given key '%s' doesn't match any of the supported key %s" % (key, self.keys))
        self.cached[key].add(value)
        self.current[key].add(value)

    def get_removed(self):
        """
        This should be called after processing. Generates (key, value) to remove.
        """
        cache = caches.get(self.cache_name)

        for key in self.keys:
            to_remove_values = self.previous[key] - self.current[key]
            for to_remove in to_remove_values:
                yield key, to_remove
            cache.set(key, self.current[key])

    @staticmethod
    def close():
        """
        Unregister and close the underlying cache.
        """
        caches.unregister("traefik.extra-services", callback=lambda c: c.close())



# -*- coding: utf-8 -*-
import os

from git import Repo, InvalidGitRepositoryError

from ddb.action import Action
from ddb.config import config
from ddb.context import context
from ddb.event import events
from ddb.utils.file import chmod


class FixFilePermissionsAction(Action):
    """
    Update file access permissions based on git index
    """

    @property
    def name(self) -> str:
        return "git:fix-file-permissions"

    @property
    def event_bindings(self):
        return events.phase.configure

    @property
    def disabled(self) -> bool:
        return not config.data.get("git.fix_files_permissions") or os.name == 'nt'

    def execute(self):
        """
        Execute the action
        :return:
        """
        try:
            repo = Repo(config.paths.project_home)
            self.process_repository(repo)
        except InvalidGitRepositoryError:
            pass

    def process_repository(self, repo: Repo):
        """
        Process a repository
        :param repo: the repository to process
        :return:
        """
        for file_data in repo.git.ls_files(s=True).splitlines():
            file_config, file_path = file_data.split("\t")
            git_mode = file_config.split(" ")[0]
            file_full_path = os.path.join(repo.working_dir, file_path)
            FixFilePermissionsAction.update_mode(file_full_path, git_mode)

        for submodule in repo.submodules:
            self.process_repository(submodule.module())

    @staticmethod
    def update_mode(path, git_mode):
        """
        Update the chmod of the file
        :param path: the path to the file
        :param git_mode: the chmod to apply
        :return:
        """
        if os.path.isdir(path):
            return

        if git_mode[-4:] == '0755' and not os.access(path, os.X_OK):
            chmod(path, '+x')
            context.log.warning('If this chmod is not expected, update the file permission in git using command "git '
                                'update-index --chmod=+x foo.sh"')
            return

        if git_mode[-4:] != '0755' and os.access(path, os.X_OK):
            chmod(path, '-x')
            context.log.warning('If this chmod is not expected, update the file permission in git using command "git '
                                'update-index --chmod=+x foo.sh"')

    @staticmethod
    def get_current_chmod(path: str) -> str:
        """
        Retrieve the current chmod for the given path
        :param path: the path to check
        :return:
        """
        return oct(os.lstat(path).st_mode)[-6::]



# -*- coding: utf-8 -*-

from marshmallow import fields, Schema
from marshmallow_union import Union

from ddb.feature.schema import FeatureSchema, DisableableSchema


class RegistrySchema(Schema):
    """
    Registry schema
    """
    name = fields.String(required=False, allow_none=True, default=None)
    repository = fields.String(required=False, allow_none=True, default=None)


class BuildContextSchema(Schema):
    """
    Build context schema
    """
    base_directory = fields.String(required=True, default=".docker")
    use_project_home = fields.Boolean(required=False, default=False)


class BuildSchema(DisableableSchema):
    """
    Build schema
    """
    cache_from_image = fields.Boolean(required=True, default=False)
    context = fields.Nested(BuildContextSchema(), default=BuildContextSchema())
    image_tag = fields.String(required=False, allow_none=True,
                              default=None)  # default is set in feature _configure_defaults
    image_tag_from = Union(fields=[fields.Boolean(), fields.String()],
                           required=False, allow_none=True, default=False)


class BinarySchema(DisableableSchema):
    """
    Binary schema
    """
    global_ = fields.Boolean(data_key='global', attribute='global', required=False, allow_none=True, default=None)


class VirtualhostSchema(DisableableSchema):  # Replacement for docker.reverse-proxy schema
    """
    VirtualHost schema
    """
    type = fields.String(required=True, default="traefik")
    network_id = fields.String(required=True, default="reverse-proxy")
    certresolver = fields.String(required=False, allow_none=True, default=None)
    https = fields.Boolean(required=True, default=True)
    redirect_to_https = fields.Boolean(required=False, allow_none=True, default=None)
    redirect_to_path_prefix = fields.Boolean(required=False, allow_none=True, default=None)


class XDebugSchema(DisableableSchema):
    """
    XDebug schema
    """
    host = fields.String(required=True, allow_none=True, default=None)  # default is set in feature _configure_defaults
    mode = fields.String(required=False, default="debug")
    version = fields.String(required=False, allow_none=True, default=None)
    session = fields.String(required=False, allow_none=True,
                            default=None)  # default is set in feature _configure_defaults


class ServiceSchema(DisableableSchema):
    """
    Service schema
    """
    init = fields.Boolean(required=False, default=True)
    restart = fields.String(required=False, allow_none=True, default=None)


class ExposeSchema(DisableableSchema):
    """
    Expose schema
    """
    port_prefix = fields.Integer(required=False)  # default is set in feature _configure_defaults


class MountSchema(DisableableSchema):
    """
    Mount schema
    """
    directory = fields.String(required=False, allow_none=True, default=None)
    directories = fields.Dict(required=False, allow_none=True, default=None,
                              keys=fields.String(), values=fields.String())


class UserSchema(DisableableSchema):
    """
    User schema
    """
    uid = fields.Integer(required=True, default=None)  # default is set in feature _configure_defaults
    gid = fields.Integer(required=True, default=None)  # default is set in feature _configure_defaults
    name = fields.String(required=False, allow_none=True, default=None)  # default is set in feature _configure_defaults
    group = fields.String(required=False, allow_none=True,
                          default=None)  # default is set in feature _configure_defaults
    name_to_uid = fields.Dict(required=True, default={})
    group_to_gid = fields.Dict(required=True, default={})


class ComposeSchema(Schema):
    """
    Docker compose schema
    """
    project_name = fields.String(required=True, default=None)  # default is set in feature _configure_defaults
    network_name = fields.String(required=True, default=None)  # default is set in feature _configure_defaults
    version = fields.String(required=True, default="3.7")
    excluded_services = fields.List(fields.String(), required=False, allow_none=True, default=[])
    included_services = fields.List(fields.String(), required=False, allow_none=True, default=None)


class NetworksSchema(Schema):
    """
    Networks schema.
    """
    names = fields.Dict(required=True, default={})


class DockerSchema(Schema):
    """
    Docker schema
    """
    compose = fields.Nested(ComposeSchema(), default=ComposeSchema())
    networks = fields.Nested(NetworksSchema(), default=NetworksSchema())
    build = fields.Nested(BuildSchema(), default=BuildSchema())
    service = fields.Nested(ServiceSchema(), default=ServiceSchema())
    expose = fields.Nested(ExposeSchema(), default=ExposeSchema())
    mount = fields.Nested(MountSchema(), default=MountSchema())
    registry = fields.Nested(RegistrySchema(), default=RegistrySchema())
    user = fields.Nested(UserSchema(), default=UserSchema())
    binary = fields.Nested(BinarySchema(), default=BinarySchema())
    virtualhost = fields.Nested(VirtualhostSchema(), default=VirtualhostSchema())
    xdebug = fields.Nested(XDebugSchema(), default=XDebugSchema())


class JsonnetSchema(FeatureSchema):
    """
    Jsonnet schema.
    """
    suffixes = fields.List(fields.String(), default=[".jsonnet"])
    extensions = fields.List(fields.String(), default=[".*", ""])
    includes = fields.List(fields.String())  # default is build automatically from suffixes value
    excludes = fields.List(fields.String())
    docker = fields.Nested(DockerSchema(), default=DockerSchema())



import os

from ddb.__main__ import main


class TestPlugin:
    def test_copy_ca_certificates(self, project_loader):
        project_loader("feature")

        assert not os.path.exists("test")
        assert not os.path.exists("test2")
        assert not os.path.exists("test3")
        assert not os.path.exists("some")

        main(["configure"])

        assert os.path.exists("test")
        assert os.path.exists("test2")
        assert os.path.exists("test3")
        assert os.path.exists("some")



import os
import sys

import yaml

from ddb.utils.docker import DockerUtils, DockerComposeYamlMissingException


class TestDockerUtils:
    def test_get_config(self, data_dir: str):
        os.chdir(data_dir)

        yaml_config = DockerUtils.get_config()

        assert os.path.exists('docker-compose.yml')
        with open('docker-compose.yml', 'r') as f:
            expected = yaml.load(f.read(), yaml.SafeLoader)

        assert expected == yaml_config

    def test_basic_manipulations(self, data_dir: str):
        container_name = 'web'
        os.chdir(data_dir)
        DockerUtils.service_stop(container_name)

        assert not DockerUtils.is_container_up(container_name)

        DockerUtils.service_up(container_name)

        assert DockerUtils.is_container_up(container_name)
        DockerUtils.service_stop(container_name)

        assert not DockerUtils.is_container_up(container_name)



# -*- coding: utf-8 -*-
from abc import abstractmethod, ABC
from typing import Iterable, Tuple

from ddb.registry import RegistryObject


class Binary(RegistryObject, ABC):
    """
    Wraps a binary on the system.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Name of the binary.
        """

    @abstractmethod
    def command(self, *args) -> Iterable[str]:
        """
        Get the binary command.
        """

    @abstractmethod
    def should_run(self, *args) -> bool:
        """
        Check if this binary should run.
        """

    @abstractmethod
    def priority(self) -> int:
        """
        Priority of this binary amoung binaries with the same name that should run.
        """

    @abstractmethod
    def before_run(self, *args) -> None:
        """
        Add action to be executed before running the command.
        """

    @property
    @abstractmethod
    def global_(self) -> bool:
        """
        Check if binary should be registered globally.
        """

    @abstractmethod
    def __eq__(self, other) -> bool:
        pass

    @abstractmethod
    def __hash__(self):
        pass


class AbstractBinary(Binary, ABC):
    """
    Abstract implementation for binary.
    """

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def before_run(self, *args) -> None:
        return None

    def should_run(self, *args) -> bool:
        return True

    def priority(self) -> int:
        return 0

    @property
    def global_(self) -> bool:
        return False

    def __eq__(self, other) -> bool:
        if not isinstance(other, AbstractBinary):
            return False
        return self.name == other.name

    def __hash__(self):
        return hash((self.name,))


class DefaultBinary(AbstractBinary):
    """
    Default implementation for binary.
    """

    def __init__(self, name: str, command: Iterable[str]):
        super().__init__(name)
        self._command = tuple(command)

    def command(self, *args) -> Tuple[str]:
        return self._command

    def __eq__(self, other) -> bool:
        if not super().__eq__(other):
            return False
        if not isinstance(other, DefaultBinary):
            return False
        return self.command() == other.command()

    def __hash__(self):
        return hash((super().__hash__(), self.command()))



from typing import Optional, Union, Iterable

from ddb.binary import Binary as BinaryType
from . import event
# pylint:disable=redefined-outer-name
from ..command import Command


class File:
    """
    File related events.
    """

    @event("file:deleted")
    def deleted(self, file: str):
        """
        When a file is deleted.
        :param file:
        """

    @event("file:found")
    def found(self, file: str):
        """
        When a file is found.
        :param file:
        """

    @event("file:generated")
    def generated(self, source: Optional[str], target: str):
        """
        When a file is generated.
        :param source:
        :param target:
        """


class Certs:
    """
    Certificates related events.
    """

    @event("certs:generate")
    def generate(self, domain: str, wildcard=True):
        """
        Ask to generate a certificate.
        :param domain:
        :param wildcard:
        """

    @event("certs:remove")
    def remove(self, domain: str):
        """
        Ask to remove a certificate.
        :param domain:
        """

    @event("certs:generated")
    def generated(self, domain: str, wildcard: bool, private_key: Union[bytes, str], certificate: Union[bytes, str]):
        """
        When a certificate is generated.
        :param domain:
        :param wildcard:
        :param private_key:
        :param certificate:
        """

    @event("certs:removed")
    def removed(self, domain: str):
        """
        When a certificate is removed.
        :param domain:
        """

    @event("certs:available")
    def available(self, domain: str, wildcard: bool, private_key: Union[bytes, str], certificate: Union[bytes, str]):
        """
        When a certificate is available.
        :param domain:
        :param wildcard:
        :param private_key:
        :param certificate:
        :return:
        """


class Binary:
    """
    Binary related events.
    """

    @event("binary:unregistered")
    def unregistered(self, binary: BinaryType):
        """
        When a binary as been unregistered.
        :param binary:
        :return:
        """

    @event("binary:registered")
    def registered(self, binary: BinaryType):
        """
        When a binary has been registered.
        :param binary:
        :return:
        """

    @event("binary:found")
    def found(self, binary: BinaryType):
        """
        When a binary has been found.
        :param binary:
        """


class Run:
    """
    Run related events.
    """

    @event("run:run")
    def run(self, name: str):
        """
        When a command has been runned.
        :param name:
        """

    @event("run:command")
    def command(self, command: Iterable[str], system_path: bool = False):
        """
        When a command request should be handled.
        :param command:
        :param system_path:
        """


class Docker:
    """
    Docker related events.
    """

    @event("docker:docker-compose-config")
    def docker_compose_config(self, docker_compose_config: dict):
        """
        When a docker compose configuration has been read.
        :param docker_compose_config:
        """

    @event("docker:docker-compose-before-events")
    def docker_compose_before_events(self, docker_compose_config: dict):
        """
        Before docker compose to emit events.
        :param docker_compose_config:
        """

    @event("docker:docker-compose-after-events")
    def docker_compose_after_events(self, docker_compose_config: dict):
        """
        After docker compose has emitted events.
        :param docker_compose_config:
        """

    @event("docker:binary")
    def binary(self, name=None, workdir=None, options=None, options_condition=None, condition=None, args=None,
               exe=False, entrypoint=None, global_=None, docker_compose_service=None):
        """
        Ask a binary to be generated from docker compose service
        """


class Phase:
    """
    Phase related events.
    """

    @event("phase:init")
    def init(self):
        """
        Init phase
        """

    @event("phase:configure")
    def configure(self):
        """
        Configure phase
        """

    @event("phase:download")
    def download(self):
        """
        Download phase
        """

    @event("phase:features")
    def features(self):
        """
        Features phase
        """

    @event("phase:info")
    def info(self):
        """
        Features phase
        """

    @event("phase:config")
    def config(self):
        """
        Config phase
        """

    @event("phase:run")
    def run(self):
        """
        Run phase
        """

    @event("phase:selfupdate")
    def selfupdate(self):
        """
        Selfupdate phase
        """


class Config:
    """
    Phase related events.
    """

    @event("config:reloaded")
    def reloaded(self):
        """
        Config has been reloaded
        """


class Main:
    """
    Main function related events.
    """

    @event("main:start")
    def start(self, command: Command):
        """
        Main start
        """

    @event("main:terminate")
    def terminate(self, command: Command):
        """
        Main terminate
        """

    @event("main:version")
    def version(self, silent: bool):
        """
        Main version
        """


file = File()
certs = Certs()
binary = Binary()
run = Run()
docker = Docker()
phase = Phase()
config = Config()
main = Main()
